###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

"""
    This test is of the core IDAES components that allow for the declaration
    and usage of solids phases in conjunction with aqueous phases.

    In these tests, we assess the convergence behavior of the solubility problem
    under various pH conditions for a solubility reaction that is inherently pH
    dependent. Several cases will be tested under different pH conditions and
    with different levels of complexity.

    Case 1 (remineralization, post-RO):
        Aqueous Rxns:   H2O <--> H + OH             logK = -14
        Solubility:     Ca(OH)2 <--> Ca + 2 OH      logK = -5.26


    Case 2 (softening, without explicit lime):
        Aqueous Rxns:   H2O <--> H + OH             logK = -14
                        H2CO3 <--> H + HCO3         logK = -6.3
                        HCO3 <--> H + CO3           logK = -10.2
        Solubility:     CaCO3 <--> Ca + CO3         logK = -12


    [NOTE: IPOPT/IDAES does not handle simultaneous dissolution and
            precipitation very well. To implement this in a flowsheet,
            there should be separate units for dissolution and precipitation.
            For instance, have 1 unit for addition and dissolution of lime,
            then a separate downstream unit for precipitation of CaCO3.]
    Case 3: (softening, with lime: precipitation + dissolution)
        Aqueous Rxns:   H2O <--> H + OH             logK = -14
                        H2CO3 <--> H + HCO3         logK = -6.3
                        HCO3 <--> H + CO3           logK = -10.2
        Solubility:     CaCO3 <--> Ca + CO3         logK = -12
                        Ca(OH)2 <--> Ca + 2 OH      logK = -5.26


    Case 4: (phosphorus removal, most realistic test)
        Aqueous Rxns:
                        H2O <---> H + OH
                        H2CO3 <---> H + HCO3
                        HCO3 <---> H + CO3
                        H2PO4 <---> H + HPO4
                        HPO4 <---> H + PO4
                        FeOH <---> Fe + OH
                        Fe(OH)2 <---> FeOH + OH
                        Fe(OH)3 <---> Fe(OH)2 + OH
                        Fe(OH)4 <---> Fe(OH)3 + OH
        Solubility:
                        FePO4 <---> Fe + PO4
"""

# Importing testing libraries
import pytest

# Importing the object for units from pyomo
from pyomo.environ import units as pyunits
from pyomo.environ import Var

# Imports from idaes core
from idaes.core import AqueousPhase, SolidPhase, FlowsheetBlock, EnergyBalanceType
from idaes.core.components import Solvent, Solute, Cation, Anion, Component
from idaes.core.phases import PhaseType as PT

# Imports from idaes generic models
import idaes.generic_models.properties.core.pure.Perrys as Perrys
from idaes.generic_models.properties.core.pure.ConstantProperties import Constant
from idaes.generic_models.properties.core.state_definitions import FTPx, FpcTP
from idaes.generic_models.properties.core.eos.ideal import Ideal

# Importing the enum for concentration unit basis used in the 'get_concentration_term' function
from idaes.generic_models.properties.core.generic.generic_reaction import ConcentrationForm

# Import the object/function for heat of reaction
from idaes.generic_models.properties.core.reactions.dh_rxn import constant_dh_rxn

# Import safe log power law equation
from idaes.generic_models.properties.core.reactions.equilibrium_forms import log_power_law_equil, power_law_equil

# Import built-in van't Hoff function
from idaes.generic_models.properties.core.reactions.equilibrium_constant import van_t_hoff

from idaes.generic_models.properties.core.reactions.equilibrium_forms import \
    solubility_product, log_solubility_product, log_power_law_equil
from idaes.generic_models.properties.core.reactions.equilibrium_constant import \
    ConstantKeq

# Import specific pyomo objects
from pyomo.environ import (ConcreteModel,
                           SolverStatus,
                           TerminationCondition,
                           value,
                           Suffix)

from idaes.core.util import scaling as iscale
from idaes.core.util.initialization import fix_state_vars, revert_state_vars

import idaes.logger as idaeslog

# Import pyomo methods to check the system units
from pyomo.util.check_units import assert_units_consistent

# Import idaes methods to check the model during construction
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import (degrees_of_freedom,
                                              fixed_variables_set,
                                              activated_constraints_set,
                                              number_variables,
                                              number_total_constraints,
                                              number_unused_variables)

# Import the idaes objects for Generic Properties and Reactions
from idaes.generic_models.properties.core.generic.generic_property import (
        GenericParameterBlock)
from idaes.generic_models.properties.core.generic.generic_reaction import (
        GenericReactionParameterBlock)

# Import the idaes object for the EquilibriumReactor unit model
from idaes.generic_models.unit_models.equilibrium_reactor import EquilibriumReactor

# Import log10 function from pyomo
from pyomo.environ import log10

__author__ = "Austin Ladshaw"

# Case 1 Config
case1_thermo_config = {
    "components": {
        'H2O': {"type": Solvent, "valid_phase_types": PT.aqueousPhase,
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (18.0153, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-285, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (69, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'H_+': {"type": Cation, "charge": 1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (1.00784, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (0, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (0, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'OH_-': {"type": Anion, "charge": -1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (17.008, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-230, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (-10, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'Ca_2+': {"type": Cation, "charge": 2,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (40.078, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (167039, pyunits.J/pyunits.kmol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-542.83, pyunits.J/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (-53, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'Ca(OH)2': {"type": Component, "valid_phase_types": PT.solidPhase,
              "dens_mol_sol_comp": Constant,
              "enth_mol_sol_comp": Constant,
              "cp_mol_sol_comp": Constant,
              "entr_mol_sol_comp": Constant,
              "parameter_data": {
                    "mw": (74.093, pyunits.g/pyunits.mol),
                    "dens_mol_sol_comp_coeff": (55, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_sol_comp_coeff": (167039, pyunits.J/pyunits.kmol/pyunits.K),
                    "enth_mol_form_sol_comp_ref": (-986, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_sol_comp_ref": (83, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    },

              },
              # End Component list
        "phases":  {'Liq': {"type": AqueousPhase,
                            "equation_of_state": Ideal},
                    'Sol': {"type": SolidPhase,
                            "equation_of_state": Ideal}
                    },

        "state_definition": FpcTP,
        "state_bounds": {
                         "temperature": (273.15, 300, 650),
                         "pressure": (5e4, 1e5, 1e6)
                     },

        "pressure_ref": 1e5,
        "temperature_ref": 300,
        "base_units": {"time": pyunits.s,
                       "length": pyunits.m,
                       "mass": pyunits.kg,
                       "amount": pyunits.mol,
                       "temperature": pyunits.K},

    }
    # End thermo_config definition

# Case 1 rxn config (with log_solubility_product)
case1_log_rxn_config = {
    "base_units": {"time": pyunits.s,
                   "length": pyunits.m,
                   "mass": pyunits.kg,
                   "amount": pyunits.mol,
                   "temperature": pyunits.K},
    "equilibrium_reactions": {
        "CaOH2_Ksp": {
                    "stoichiometry": {  ("Sol", "Ca(OH)2"): -1,
                                        ("Liq", "Ca_2+"): 1,
                                        ("Liq", "OH_-"): 2},
                    "heat_of_reaction": constant_dh_rxn,
                    "equilibrium_constant": van_t_hoff,
                    "equilibrium_form": log_solubility_product,
                    "concentration_form": ConcentrationForm.moleFraction,
                    "parameter_data": {
                        "dh_rxn_ref": (0.0, pyunits.J/pyunits.mol),
                        "k_eq_ref": (10**-5.26/55.2/55.2/55.2, pyunits.dimensionless),
                        "T_eq_ref": (298.0, pyunits.K),
                        "reaction_order": { ("Sol", "Ca(OH)2"): 0,
                                            ("Liq", "Ca_2+"): 1,
                                            ("Liq", "OH_-"): 2}
                        }
                        # End parameter_data
                },
        "H2O_Kw": {
                "stoichiometry": {("Liq", "H2O"): -1,
                                 ("Liq", "H_+"): 1,
                                 ("Liq", "OH_-"): 1},
               "heat_of_reaction": constant_dh_rxn,
               "equilibrium_constant": van_t_hoff,
               "equilibrium_form": log_power_law_equil,
               "concentration_form": ConcentrationForm.moleFraction,
               "parameter_data": {
                   "dh_rxn_ref": (55.830, pyunits.J/pyunits.mol),
                   "k_eq_ref": (10**-14/55.2/55.2, pyunits.dimensionless),
                   "T_eq_ref": (298, pyunits.K),

                   # By default, reaction orders follow stoichiometry
                   #    manually set reaction order here to override
                   "reaction_order": {("Liq", "H2O"): 0,
                                    ("Liq", "H_+"): 1,
                                    ("Liq", "OH_-"): 1}
                    }
                    # End parameter_data
               }
                # End Reactions
         }
         # End equilibrium_reactions
    }
    # End reaction_config definition

# Get default solver for testing
solver = get_solver()

## Set of helper functions for scaling
def _set_eps_vals(rxn_params, rxn_config):
    # For solubility reactions, have to set the eps value
    try:
        for rid in rxn_params.equilibrium_reaction_idx:
            scale = rxn_config["equilibrium_reactions"][rid]["parameter_data"]["k_eq_ref"][0]

            # NOTE: The solubility_product function STILL has an eps value that we need to set
            try:
                # Want to set eps in some fashion similar to this
                if scale < 1e-16:
                    rxn_params.component("reaction_"+rid).eps.value = scale*1e-2
                else:
                    rxn_params.component("reaction_"+rid).eps.value = 1e-16*1e-2
            except:
                pass
    except:
        pass

def _set_equ_rxn_scaling(unit, rxn_config):
    #Add scaling factors for reactions (changes depending on if it is a log form or not)
    min_scale = 1e-3
    for i in unit.control_volume.equilibrium_reaction_extent_index:
        # i[0] = time, i[1] = reaction

        scale = max(min_scale, rxn_config["equilibrium_reactions"][i[1]]["parameter_data"]["k_eq_ref"][0])

        # this may also need to be different for solubility_product
        if rxn_config["equilibrium_reactions"][i[1]]["equilibrium_form"] == solubility_product:
            iscale.set_scaling_factor(unit.control_volume.equilibrium_reaction_extent[0.0,i[1]], 1e-1/scale)
            iscale.constraint_scaling_transform(
                unit.control_volume.reactions[0.0].equilibrium_constraint[i[1]], 1e-1/scale)
        else:
            iscale.set_scaling_factor(unit.control_volume.equilibrium_reaction_extent[0.0,i[1]], 10/scale)
            iscale.constraint_scaling_transform(
                unit.control_volume.reactions[0.0].equilibrium_constraint[i[1]], 0.1)

def _set_mat_bal_scaling_FpcTP(unit):
    # For species
    min = 1e-3
    for i in unit.control_volume.properties_out[0.0].mole_frac_phase_comp:
        # i[0] = phase, i[1] = species
        if unit.inlet.flow_mol_phase_comp[0, i[0], i[1]].value > min:
            scale = unit.inlet.flow_mol_phase_comp[0, i[0], i[1]].value
        else:
            scale = min

        ## For now, these are both set the same, but may change in testing
        # Scaling factors for liquid
        if i[0] == 'Liq':
            iscale.set_scaling_factor(unit.control_volume.properties_out[0.0].mole_frac_comp[i[1]], 10/scale)
            iscale.set_scaling_factor(unit.control_volume.properties_out[0.0].mole_frac_phase_comp[i], 10/scale)
            iscale.set_scaling_factor(unit.control_volume.properties_out[0.0].flow_mol_phase_comp[i], 10/scale)
            iscale.constraint_scaling_transform(unit.control_volume.material_balances[0.0,i[1]], 10/scale)
        #Scaling factors for solids
        else:
            iscale.set_scaling_factor(unit.control_volume.properties_out[0.0].mole_frac_comp[i[1]], 10/scale)
            iscale.set_scaling_factor(unit.control_volume.properties_out[0.0].mole_frac_phase_comp[i], 10/scale)
            iscale.set_scaling_factor(unit.control_volume.properties_out[0.0].flow_mol_phase_comp[i], 10/scale)
            iscale.constraint_scaling_transform(unit.control_volume.material_balances[0.0,i[1]], 10/scale)

def _set_ene_bal_scaling(unit):
    max = 1
    min = 1
    for phase in unit.control_volume.properties_in[0.0].enth_mol_phase:
        val = abs(value(unit.control_volume.properties_in[0.0].enth_mol_phase[phase].expr))
        if val >= max:
            max = val
        if val <= min:
            val = min
        iscale.set_scaling_factor(unit.control_volume.properties_in[0.0]._enthalpy_flow_term[phase], 10/val)
        iscale.set_scaling_factor(unit.control_volume.properties_out[0.0]._enthalpy_flow_term[phase], 10/val)

    iscale.constraint_scaling_transform(unit.control_volume.enthalpy_balances[0.0], 10/max)

# Defaults to pH of 7 with no lime added
def run_case1(xOH=1e-7/55.2, xH=1e-7/55.2, xCaOH2=1e-20, xCa=1e-20,
            thermo_config=None, rxn_config=None, has_energy_balance=True):
    print("==========================================================================")
    print("Case 1: Remineralization via lime dissolution")
    print("xOH = "+str(xOH))
    print("xH = "+str(xH))
    print("xCaOH2 = "+str(xCaOH2))
    print("Initial pH = " +str(-log10(xH*55.2)))
    print()

    model = ConcreteModel()
    model.fs = FlowsheetBlock(default={"dynamic": False})
    model.fs.thermo_params = GenericParameterBlock(default=thermo_config)

    model.fs.rxn_params = GenericReactionParameterBlock(
            default={"property_package": model.fs.thermo_params,
                    **rxn_config
                    })

    args = {"property_package": model.fs.thermo_params,
            "reaction_package": model.fs.rxn_params,
            "has_rate_reactions": False,
            "has_equilibrium_reactions": True,
            "has_heat_transfer": False,
            "has_heat_of_reaction": False,
            "has_pressure_change": False}
    if has_energy_balance == False:
        args["energy_balance_type"] = EnergyBalanceType.none

    model.fs.unit = EquilibriumReactor(default=args)

    total_flow_mol = 10

    # Set flow_mol_phase_comp
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Ca_2+"].fix( xCa*total_flow_mol )

    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H_+"].fix( xH*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "OH_-"].fix( xOH*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Sol", "Ca(OH)2"].fix( xCaOH2*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix( (1-xH-xOH-xCaOH2-xCa)*total_flow_mol )

    model.fs.unit.inlet.pressure.fix(101325.0)
    model.fs.unit.inlet.temperature.fix(298.)
    if has_energy_balance == False:
        model.fs.unit.outlet.temperature.fix(298.)

    assert (degrees_of_freedom(model) == 0)

    ## ==================== Start Scaling for this problem ===========================
    _set_eps_vals(model.fs.rxn_params, rxn_config)
    _set_equ_rxn_scaling(model.fs.unit, rxn_config)
    _set_mat_bal_scaling_FpcTP(model.fs.unit)
    if has_energy_balance == True:
        _set_ene_bal_scaling(model.fs.unit)

    iscale.calculate_scaling_factors(model.fs.unit)
    assert isinstance(model.fs.unit.control_volume.scaling_factor, Suffix)
    assert isinstance(model.fs.unit.control_volume.properties_out[0.0].scaling_factor, Suffix)
    assert isinstance(model.fs.unit.control_volume.properties_in[0.0].scaling_factor, Suffix)

    ## ==================== END Scaling for this problem ===========================

    model.fs.unit.initialize(optarg=solver.options, outlvl=idaeslog.DEBUG)

    assert degrees_of_freedom(model) == 0

    results = solver.solve(model, tee=True)

    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    print("comp\toutlet.tot_molfrac")
    for i in model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp:
        print(str(i)+"\t"+str(value( model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp[i] )))
    print()

    # NOTE: Changed all to mole fraction
    for i in model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp:
        print(str(i)+"\t"+str(value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp[i])))
    print()

    Ca = value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp["Liq","Ca_2+"])
    OH = value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp["Liq","OH_-"])
    H = value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp["Liq","H_+"])
    Ksp = value(model.fs.unit.control_volume.reactions[0.0].k_eq["CaOH2_Ksp"].expr)

    print("Final pH = " +str(-log10(H*55.2)))
    print("Expected max/min pH = " +str(14+log10(xCaOH2*55.2*2)))

    print()
    if Ksp*1.01 >= Ca*OH*OH:
        print("Constraint is satisfied!")
    else:
        print("Constraint is VIOLATED!")
        print("\tRelative error: "+str(Ksp/Ca/OH/OH)+">=1")
        assert False
    print("Ksp =\t"+str(Ksp))
    print("Ca*OH**2 =\t"+str(Ca*OH*OH))

    print("==========================================================================")

    return model

## ================================= Case 1 Tests ===============================
@pytest.mark.component
def test_case_1_no_dissolution():
    model = run_case1(xOH=1e-7/55.2, xH=1e-7/55.2, xCaOH2=1e-20,
                        thermo_config=case1_thermo_config, rxn_config=case1_log_rxn_config,
                        has_energy_balance=True)

@pytest.mark.component
def test_case_1_high_dissolution():
    model = run_case1(xOH=1e-7/55.2, xH=1e-7/55.2, xCaOH2=1e-5,
                        thermo_config=case1_thermo_config, rxn_config=case1_log_rxn_config,
                        has_energy_balance=True)

@pytest.mark.component
def test_case_1_mid_dissolution():
    model = run_case1(xOH=1e-7/55.2, xH=1e-7/55.2, xCaOH2=1e-7,
                        thermo_config=case1_thermo_config, rxn_config=case1_log_rxn_config,
                        has_energy_balance=True)

@pytest.mark.component
def test_case_1_low_dissolution():
    model = run_case1(xOH=1e-7/55.2, xH=1e-7/55.2, xCaOH2=1e-9,
                        thermo_config=case1_thermo_config, rxn_config=case1_log_rxn_config,
                        has_energy_balance=True)

@pytest.mark.component
def test_case_1_high_precipitation():
    model = run_case1(xOH=1e-1/55.2, xH=1e-13/55.2, xCaOH2=1e-20, xCa=1e-1,
                        thermo_config=case1_thermo_config, rxn_config=case1_log_rxn_config,
                        has_energy_balance=True)

@pytest.mark.component
def test_case_1_low_precipitation():
    model = run_case1(xOH=1e-3/55.2, xH=1e-11/55.2, xCaOH2=1e-20, xCa=1e-1,
                        thermo_config=case1_thermo_config, rxn_config=case1_log_rxn_config,
                        has_energy_balance=True)


# Case 2 Config
'''
    Case 2 (softening, without explicit lime):
        Aqueous Rxns:   H2O <--> H + OH             logK = -14
                        H2CO3 <--> H + HCO3         logK = -6.3
                        HCO3 <--> H + CO3           logK = -10.2
        Solubility:     CaCO3 <--> Ca + CO3         logK = -12
'''
case2_thermo_config = {
    "components": {
        'H2O': {"type": Solvent, "valid_phase_types": PT.aqueousPhase,
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (18.0153, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-285, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (69, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'H_+': {"type": Cation, "charge": 1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (1.00784, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (0, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (0, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'OH_-': {"type": Anion, "charge": -1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (17.008, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-230, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (-10, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'Ca_2+': {"type": Cation, "charge": 2,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (40.078, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (167039, pyunits.J/pyunits.kmol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-542.83, pyunits.J/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (-53, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'H2CO3': {"type": Solute, "valid_phase_types": PT.aqueousPhase,
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (62.03, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-699.7, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (187, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'HCO3_-': {"type": Anion, "charge": -1,
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (62.03, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-692, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (91.2, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'CO3_2-': {"type": Anion, "charge": -2,
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (62.03, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-677.1, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (-56.9, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'CaCO3': {"type": Component, "valid_phase_types": PT.solidPhase,
              "dens_mol_sol_comp": Constant,
              "enth_mol_sol_comp": Constant,
              "cp_mol_sol_comp": Constant,
              "entr_mol_sol_comp": Constant,
              "parameter_data": {
                    "mw": (100.0869, pyunits.g/pyunits.mol),
                    "dens_mol_sol_comp_coeff": (55, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_sol_comp_coeff": (167039, pyunits.J/pyunits.kmol/pyunits.K),
                    "enth_mol_form_sol_comp_ref": (-1207, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_sol_comp_ref": (91.7, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    },

              },
              # End Component list
        "phases":  {'Liq': {"type": AqueousPhase,
                            "equation_of_state": Ideal},
                    'Sol': {"type": SolidPhase,
                            "equation_of_state": Ideal}
                    },

        "state_definition": FpcTP,
        "state_bounds": {
                         "temperature": (273.15, 300, 650),
                         "pressure": (5e4, 1e5, 1e6)
                     },

        "pressure_ref": 1e5,
        "temperature_ref": 300,
        "base_units": {"time": pyunits.s,
                       "length": pyunits.m,
                       "mass": pyunits.kg,
                       "amount": pyunits.mol,
                       "temperature": pyunits.K},

    }
    # End thermo_config definition

# Case 2 rxn config (with log_solubility_product)
case2_log_rxn_config = {
    "base_units": {"time": pyunits.s,
                   "length": pyunits.m,
                   "mass": pyunits.kg,
                   "amount": pyunits.mol,
                   "temperature": pyunits.K},
    "equilibrium_reactions": {
        "CaCO3_Ksp": {
                    "stoichiometry": {  ("Sol", "CaCO3"): -1,
                                        ("Liq", "Ca_2+"): 1,
                                        ("Liq", "CO3_2-"): 1},
                    "heat_of_reaction": constant_dh_rxn,
                    "equilibrium_constant": van_t_hoff,
                    "equilibrium_form": log_solubility_product,
                    "concentration_form": ConcentrationForm.moleFraction,
                    "parameter_data": {
                        "dh_rxn_ref": (0.0, pyunits.J/pyunits.mol),
                        "k_eq_ref": (10**-12/55.2/55.2, pyunits.dimensionless),
                        "T_eq_ref": (298.0, pyunits.K),
                        "reaction_order": { ("Sol", "CaCO3"): 0,
                                            ("Liq", "Ca_2+"): 1,
                                            ("Liq", "CO3_2-"): 1}
                        }
                        # End parameter_data
                },
        "H2O_Kw": {
                "stoichiometry": {("Liq", "H2O"): -1,
                                 ("Liq", "H_+"): 1,
                                 ("Liq", "OH_-"): 1},
               "heat_of_reaction": constant_dh_rxn,
               "equilibrium_constant": van_t_hoff,
               "equilibrium_form": log_power_law_equil,
               "concentration_form": ConcentrationForm.moleFraction,
               "parameter_data": {
                   "dh_rxn_ref": (55.830, pyunits.J/pyunits.mol),
                   "k_eq_ref": (10**-14/55.2/55.2, pyunits.dimensionless),
                   "T_eq_ref": (298, pyunits.K),

                   # By default, reaction orders follow stoichiometry
                   #    manually set reaction order here to override
                   "reaction_order": {("Liq", "H2O"): 0,
                                    ("Liq", "H_+"): 1,
                                    ("Liq", "OH_-"): 1}
                    }
                    # End parameter_data
               },
        "H2CO3_Ka1": {
                "stoichiometry": {
                    ("Liq", "H2CO3"): -1,
                    ("Liq", "H_+"): 1,
                    ("Liq", "HCO3_-"): 1,
                },
                "heat_of_reaction": constant_dh_rxn,
                "equilibrium_constant": van_t_hoff,
                "equilibrium_form": log_power_law_equil,
                "concentration_form": ConcentrationForm.moleFraction,
                "parameter_data": {
                    "dh_rxn_ref": (7.7, pyunits.kJ/pyunits.mol),
                    "k_eq_ref": (10**-6.35/55.2, pyunits.dimensionless),
                    "T_eq_ref": (298, pyunits.K),
                    # By default, reaction orders follow stoichiometry
                    #    manually set reaction order here to override
                    "reaction_order": {
                        ("Liq", "H2CO3"): -1,
                        ("Liq", "H_+"): 1,
                        ("Liq", "HCO3_-"): 1,
                    },
                }
                # End parameter_data
            },
        "H2CO3_Ka2": {
                "stoichiometry": {
                    ("Liq", "HCO3_-"): -1,
                    ("Liq", "H_+"): 1,
                    ("Liq", "CO3_2-"): 1,
                },
                "heat_of_reaction": constant_dh_rxn,
                "equilibrium_constant": van_t_hoff,
                "equilibrium_form": log_power_law_equil,
                "concentration_form": ConcentrationForm.moleFraction,
                "parameter_data": {
                    "dh_rxn_ref": (14.9, pyunits.kJ/pyunits.mol),
                    "k_eq_ref": (10**-10.33/55.2, pyunits.dimensionless),
                    "T_eq_ref": (298, pyunits.K),
                    # By default, reaction orders follow stoichiometry
                    #    manually set reaction order here to override
                    "reaction_order": {
                        ("Liq", "HCO3_-"): -1,
                        ("Liq", "H_+"): 1,
                        ("Liq", "CO3_2-"): 1,
                    },
                }
                # End parameter_data
            }
            #End last reaction
         }
         # End equilibrium_reactions
    }
    # End reaction_config definition

# Defaults to pH of 7 with no lime added
def run_case2(xOH=1e-7/55.2, xH=1e-7/55.2, xCaCO3=1e-20, xCa=1e-20, xH2CO3=1e-20, xHCO3=1e-20, xCO3=1e-20,
            thermo_config=None, rxn_config=None, has_energy_balance=True):
    print("==========================================================================")
    print("Case 2: Water softening via pH changes")
    print("xOH = "+str(xOH))
    print("xH = "+str(xH))
    print("xCaCO3 = "+str(xCaCO3))
    print("xCa = "+str(xCa))
    print("xH2CO3 = "+str(xH2CO3))
    print("xHCO3 = "+str(xHCO3))
    print("xCO3 = "+str(xCO3))
    print()

    model = ConcreteModel()
    model.fs = FlowsheetBlock(default={"dynamic": False})
    model.fs.thermo_params = GenericParameterBlock(default=thermo_config)

    model.fs.rxn_params = GenericReactionParameterBlock(
            default={"property_package": model.fs.thermo_params,
                    **rxn_config
                    })

    args = {"property_package": model.fs.thermo_params,
            "reaction_package": model.fs.rxn_params,
            "has_rate_reactions": False,
            "has_equilibrium_reactions": True,
            "has_heat_transfer": False,
            "has_heat_of_reaction": False,
            "has_pressure_change": False}
    if has_energy_balance == False:
        args["energy_balance_type"] = EnergyBalanceType.none

    model.fs.unit = EquilibriumReactor(default=args)

    total_flow_mol = 10

    # Set flow_mol_phase_comp
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Ca_2+"].fix( xCa*total_flow_mol )

    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H_+"].fix( xH*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "OH_-"].fix( xOH*total_flow_mol )

    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2CO3"].fix( xH2CO3*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "HCO3_-"].fix( xHCO3*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "CO3_2-"].fix( xCO3*total_flow_mol )

    model.fs.unit.inlet.flow_mol_phase_comp[0, "Sol", "CaCO3"].fix( xCaCO3*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix( (1-xH-xOH-xCaCO3-xCa-xH2CO3-xHCO3-xCO3)*total_flow_mol )

    model.fs.unit.inlet.pressure.fix(101325.0)
    model.fs.unit.inlet.temperature.fix(298.)
    if has_energy_balance == False:
        model.fs.unit.outlet.temperature.fix(298.)

    assert (degrees_of_freedom(model) == 0)

    ## ==================== Start Scaling for this problem ===========================
    _set_eps_vals(model.fs.rxn_params, rxn_config)
    _set_equ_rxn_scaling(model.fs.unit, rxn_config)
    _set_mat_bal_scaling_FpcTP(model.fs.unit)
    if has_energy_balance == True:
        _set_ene_bal_scaling(model.fs.unit)

    iscale.calculate_scaling_factors(model.fs.unit)
    assert isinstance(model.fs.unit.control_volume.scaling_factor, Suffix)
    assert isinstance(model.fs.unit.control_volume.properties_out[0.0].scaling_factor, Suffix)
    assert isinstance(model.fs.unit.control_volume.properties_in[0.0].scaling_factor, Suffix)

    ## ==================== END Scaling for this problem ===========================

    model.fs.unit.initialize(optarg=solver.options, outlvl=idaeslog.DEBUG)

    assert degrees_of_freedom(model) == 0

    results = solver.solve(model, tee=True)

    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    ## ==================== Check the results ================================
    print("comp\toutlet.tot_molfrac")
    for i in model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp:
        print(str(i)+"\t"+str(value( model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp[i] )))
    print()

    # NOTE: Changed all to mole fraction
    for i in model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp:
        print(str(i)+"\t"+str(value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp[i])))
    print()

    Ca = value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp["Liq","Ca_2+"])
    OH = value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp["Liq","OH_-"])
    H = value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp["Liq","H_+"])
    CO3 = value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp["Liq","CO3_2-"])
    Ksp = value(model.fs.unit.control_volume.reactions[0.0].k_eq["CaCO3_Ksp"].expr)

    print("Final pH = " +str(-log10(H*55.2)))

    print()
    print("Ksp =\t"+str(Ksp))
    print("Ca*CO3 =\t"+str(Ca*CO3))
    print()
    if Ksp*1.01 >= Ca*CO3:
        print("Constraint is satisfied!")
    else:
        print("Constraint is VIOLATED!")
        print("\tRelative error: "+str(Ksp/Ca/CO3)+">=1")
        assert False

    print("==========================================================================")

    return model

## ================================= Case 1 Tests ===============================
@pytest.mark.component
def test_case_2_do_nothing():
    model = run_case2(xOH=1e-7/55.2, xH=1e-7/55.2, xCaCO3=1e-20, xCa=1e-20, xH2CO3=1e-20, xHCO3=1e-20, xCO3=1e-20,
                thermo_config=case2_thermo_config, rxn_config=case2_log_rxn_config, has_energy_balance=True)

@pytest.mark.component
def test_case_2_seawater_no_ca():
    model = run_case2(xOH=1e-7/55.2, xH=1e-7/55.2, xCaCO3=1e-20, xCa=1e-20,
                    xH2CO3=1e-20, xHCO3=0.00206/55.2, xCO3=1e-20,
                    thermo_config=case2_thermo_config, rxn_config=case2_log_rxn_config,
                    has_energy_balance=True)

@pytest.mark.component
def test_case_2_seawater_with_ca():
    model = run_case2(xOH=1e-7/55.2, xH=1e-7/55.2, xCaCO3=1e-20, xCa=200/50000/2/55.2,
                    xH2CO3=1e-20, xHCO3=0.00206/55.2, xCO3=1e-20,
                    thermo_config=case2_thermo_config, rxn_config=case2_log_rxn_config,
                    has_energy_balance=True)

@pytest.mark.component
def test_case_2_seawater_added_carbonates():
    model = run_case2(xOH=1e-7/55.2, xH=1e-7/55.2, xCaCO3=1e-20, xCa=200/50000/2/55.2,
                    xH2CO3=1e-20, xHCO3=0.00206/55.2, xCO3=0.1/55.2,
                    thermo_config=case2_thermo_config, rxn_config=case2_log_rxn_config,
                    has_energy_balance=True)

@pytest.mark.component
def test_case_2_low_pH_no_precip():
    model = run_case2(xOH=1e-7/55.2, xH=1e-7/55.2, xCaCO3=1e-20, xCa=200/50000/2/55.2,
                    xH2CO3=0.00206/55.2, xHCO3=1e-20, xCO3=1e-20,
                    thermo_config=case2_thermo_config, rxn_config=case2_log_rxn_config,
                    has_energy_balance=True)

@pytest.mark.component
def test_case_2_ultra_high_ca():
    model = run_case2(xOH=1e-7/55.2, xH=1e-7/55.2, xCaCO3=1e-20, xCa=6000/50000/2/55.2,
                    xH2CO3=1e-20, xHCO3=0.00206/55.2, xCO3=0.1/55.2,
                    thermo_config=case2_thermo_config, rxn_config=case2_log_rxn_config,
                    has_energy_balance=True)


# Case 3 Config
'''
    Case 3: (softening, with lime: precipitation + dissolution)
        Aqueous Rxns:   H2O <--> H + OH             logK = -14
                        H2CO3 <--> H + HCO3         logK = -6.3
                        HCO3 <--> H + CO3           logK = -10.2
        Solubility:     CaCO3 <--> Ca + CO3         logK = -12
                        Ca(OH)2 <--> Ca + 2 OH      logK = -5.26
'''
case3_thermo_config = {
    "components": {
        'H2O': {"type": Solvent, "valid_phase_types": PT.aqueousPhase,
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (18.0153, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-285, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (69, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'H_+': {"type": Cation, "charge": 1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (1.00784, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (0, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (0, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'OH_-': {"type": Anion, "charge": -1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (17.008, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-230, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (-10, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'Ca_2+': {"type": Cation, "charge": 2,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (40.078, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (167039, pyunits.J/pyunits.kmol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-542.83, pyunits.J/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (-53, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'H2CO3': {"type": Solute, "valid_phase_types": PT.aqueousPhase,
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (62.03, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-699.7, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (187, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'HCO3_-': {"type": Anion, "charge": -1,
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (62.03, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-692, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (91.2, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'CO3_2-': {"type": Anion, "charge": -2,
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (62.03, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-677.1, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (-56.9, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'CaCO3': {"type": Component, "valid_phase_types": PT.solidPhase,
              "dens_mol_sol_comp": Constant,
              "enth_mol_sol_comp": Constant,
              "cp_mol_sol_comp": Constant,
              "entr_mol_sol_comp": Constant,
              "parameter_data": {
                    "mw": (100.0869, pyunits.g/pyunits.mol),
                    "dens_mol_sol_comp_coeff": (55, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_sol_comp_coeff": (167039, pyunits.J/pyunits.kmol/pyunits.K),
                    "enth_mol_form_sol_comp_ref": (-1207, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_sol_comp_ref": (91.7, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    },
        'Ca(OH)2': {"type": Component, "valid_phase_types": PT.solidPhase,
              "dens_mol_sol_comp": Constant,
              "enth_mol_sol_comp": Constant,
              "cp_mol_sol_comp": Constant,
              "entr_mol_sol_comp": Constant,
              "parameter_data": {
                    "mw": (74.093, pyunits.g/pyunits.mol),
                    "dens_mol_sol_comp_coeff": (55, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_sol_comp_coeff": (167039, pyunits.J/pyunits.kmol/pyunits.K),
                    "enth_mol_form_sol_comp_ref": (-986, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_sol_comp_ref": (83, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    },

              },
              # End Component list
        "phases":  {'Liq': {"type": AqueousPhase,
                            "equation_of_state": Ideal},
                    'Sol': {"type": SolidPhase,
                            "equation_of_state": Ideal}
                    },

        "state_definition": FpcTP,
        "state_bounds": {
                         "temperature": (273.15, 300, 650),
                         "pressure": (5e4, 1e5, 1e6)
                     },

        "pressure_ref": 1e5,
        "temperature_ref": 300,
        "base_units": {"time": pyunits.s,
                       "length": pyunits.m,
                       "mass": pyunits.kg,
                       "amount": pyunits.mol,
                       "temperature": pyunits.K},

    }
    # End thermo_config definition

# Case 3 rxn config (with log_solubility_product)
case3_log_rxn_config = {
    "base_units": {"time": pyunits.s,
                   "length": pyunits.m,
                   "mass": pyunits.kg,
                   "amount": pyunits.mol,
                   "temperature": pyunits.K},
    "equilibrium_reactions": {
        "CaCO3_Ksp": {
                    "stoichiometry": {  ("Sol", "CaCO3"): -1,
                                        ("Liq", "Ca_2+"): 1,
                                        ("Liq", "CO3_2-"): 1},
                    "heat_of_reaction": constant_dh_rxn,
                    "equilibrium_constant": van_t_hoff,
                    "equilibrium_form": log_solubility_product,
                    "concentration_form": ConcentrationForm.moleFraction,
                    "parameter_data": {
                        "dh_rxn_ref": (0.0, pyunits.J/pyunits.mol),
                        "k_eq_ref": (10**-12/55.2/55.2, pyunits.dimensionless),
                        "T_eq_ref": (298.0, pyunits.K),
                        "reaction_order": { ("Sol", "CaCO3"): 0,
                                            ("Liq", "Ca_2+"): 1,
                                            ("Liq", "CO3_2-"): 1}
                        }
                        # End parameter_data
                },
        "CaOH2_Ksp": {
                    "stoichiometry": {  ("Sol", "Ca(OH)2"): -1,
                                        ("Liq", "Ca_2+"): 1,
                                        ("Liq", "OH_-"): 2},
                    "heat_of_reaction": constant_dh_rxn,
                    "equilibrium_constant": van_t_hoff,
                    "equilibrium_form": log_solubility_product,
                    "concentration_form": ConcentrationForm.moleFraction,
                    "parameter_data": {
                        "dh_rxn_ref": (0.0, pyunits.J/pyunits.mol),
                        "k_eq_ref": (10**-5.26/55.2/55.2/55.2, pyunits.dimensionless),
                        "T_eq_ref": (298.0, pyunits.K),
                        "reaction_order": { ("Sol", "Ca(OH)2"): 0,
                                            ("Liq", "Ca_2+"): 1,
                                            ("Liq", "OH_-"): 2}
                        }
                        # End parameter_data
                },
        "H2O_Kw": {
                "stoichiometry": {("Liq", "H2O"): -1,
                                 ("Liq", "H_+"): 1,
                                 ("Liq", "OH_-"): 1},
               "heat_of_reaction": constant_dh_rxn,
               "equilibrium_constant": van_t_hoff,
               "equilibrium_form": log_power_law_equil,
               "concentration_form": ConcentrationForm.moleFraction,
               "parameter_data": {
                   "dh_rxn_ref": (55.830, pyunits.J/pyunits.mol),
                   "k_eq_ref": (10**-14/55.2/55.2, pyunits.dimensionless),
                   "T_eq_ref": (298, pyunits.K),

                   # By default, reaction orders follow stoichiometry
                   #    manually set reaction order here to override
                   "reaction_order": {("Liq", "H2O"): 0,
                                    ("Liq", "H_+"): 1,
                                    ("Liq", "OH_-"): 1}
                    }
                    # End parameter_data
               },
        "H2CO3_Ka1": {
                "stoichiometry": {
                    ("Liq", "H2CO3"): -1,
                    ("Liq", "H_+"): 1,
                    ("Liq", "HCO3_-"): 1,
                },
                "heat_of_reaction": constant_dh_rxn,
                "equilibrium_constant": van_t_hoff,
                "equilibrium_form": log_power_law_equil,
                "concentration_form": ConcentrationForm.moleFraction,
                "parameter_data": {
                    "dh_rxn_ref": (7.7, pyunits.kJ/pyunits.mol),
                    "k_eq_ref": (10**-6.35/55.2, pyunits.dimensionless),
                    "T_eq_ref": (298, pyunits.K),
                    # By default, reaction orders follow stoichiometry
                    #    manually set reaction order here to override
                    "reaction_order": {
                        ("Liq", "H2CO3"): -1,
                        ("Liq", "H_+"): 1,
                        ("Liq", "HCO3_-"): 1,
                    },
                }
                # End parameter_data
            },
        "H2CO3_Ka2": {
                "stoichiometry": {
                    ("Liq", "HCO3_-"): -1,
                    ("Liq", "H_+"): 1,
                    ("Liq", "CO3_2-"): 1,
                },
                "heat_of_reaction": constant_dh_rxn,
                "equilibrium_constant": van_t_hoff,
                "equilibrium_form": log_power_law_equil,
                "concentration_form": ConcentrationForm.moleFraction,
                "parameter_data": {
                    "dh_rxn_ref": (14.9, pyunits.kJ/pyunits.mol),
                    "k_eq_ref": (10**-10.33/55.2, pyunits.dimensionless),
                    "T_eq_ref": (298, pyunits.K),
                    # By default, reaction orders follow stoichiometry
                    #    manually set reaction order here to override
                    "reaction_order": {
                        ("Liq", "HCO3_-"): -1,
                        ("Liq", "H_+"): 1,
                        ("Liq", "CO3_2-"): 1,
                    },
                }
                # End parameter_data
            }
            #End last reaction
         }
         # End equilibrium_reactions
    }
    # End reaction_config definition


# Defaults to pH of 7 with no lime added
def run_case3(xOH=1e-7/55.2, xH=1e-7/55.2, xCaCO3=1e-20, xCaOH2 = 1e-20, xCa=1e-20,
                xH2CO3=1e-20, xHCO3=1e-20, xCO3=1e-20,
            thermo_config=None, rxn_config=None, has_energy_balance=True):
    print("==========================================================================")
    print("Case 3: Water softening through explicit addition of lime")
    print("xOH = "+str(xOH))
    print("xH = "+str(xH))
    print("xCaCO3 = "+str(xCaCO3))
    print("xCaOH2 = "+str(xCaOH2))
    print("xCa = "+str(xCa))
    print("xH2CO3 = "+str(xH2CO3))
    print("xHCO3 = "+str(xHCO3))
    print("xCO3 = "+str(xCO3))
    print()

    model = ConcreteModel()
    model.fs = FlowsheetBlock(default={"dynamic": False})
    model.fs.thermo_params = GenericParameterBlock(default=thermo_config)

    model.fs.rxn_params = GenericReactionParameterBlock(
            default={"property_package": model.fs.thermo_params,
                    **rxn_config
                    })

    args = {"property_package": model.fs.thermo_params,
            "reaction_package": model.fs.rxn_params,
            "has_rate_reactions": False,
            "has_equilibrium_reactions": True,
            "has_heat_transfer": False,
            "has_heat_of_reaction": False,
            "has_pressure_change": False}
    if has_energy_balance == False:
        args["energy_balance_type"] = EnergyBalanceType.none

    model.fs.unit = EquilibriumReactor(default=args)

    total_flow_mol = 10

    # Set flow_mol_phase_comp
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Ca_2+"].fix( xCa*total_flow_mol )

    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H_+"].fix( xH*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "OH_-"].fix( xOH*total_flow_mol )

    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2CO3"].fix( xH2CO3*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "HCO3_-"].fix( xHCO3*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "CO3_2-"].fix( xCO3*total_flow_mol )

    model.fs.unit.inlet.flow_mol_phase_comp[0, "Sol", "CaCO3"].fix( xCaCO3*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Sol", "Ca(OH)2"].fix( xCaOH2*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix( (1-xH-xOH-xCaCO3-xCaOH2-xCa-xH2CO3-xHCO3-xCO3)*total_flow_mol )

    model.fs.unit.inlet.pressure.fix(101325.0)
    model.fs.unit.inlet.temperature.fix(298.)
    if has_energy_balance == False:
        model.fs.unit.outlet.temperature.fix(298.)

    assert (degrees_of_freedom(model) == 0)

    ## ==================== Start Scaling for this problem ===========================
    _set_eps_vals(model.fs.rxn_params, rxn_config)
    _set_equ_rxn_scaling(model.fs.unit, rxn_config)
    _set_mat_bal_scaling_FpcTP(model.fs.unit)
    if has_energy_balance == True:
        _set_ene_bal_scaling(model.fs.unit)

    iscale.calculate_scaling_factors(model.fs.unit)
    assert isinstance(model.fs.unit.control_volume.scaling_factor, Suffix)
    assert isinstance(model.fs.unit.control_volume.properties_out[0.0].scaling_factor, Suffix)
    assert isinstance(model.fs.unit.control_volume.properties_in[0.0].scaling_factor, Suffix)

    ## ==================== END Scaling for this problem ===========================

    model.fs.unit.initialize(optarg=solver.options, outlvl=idaeslog.DEBUG)

    assert degrees_of_freedom(model) == 0

    results = solver.solve(model, tee=True)

    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    ## ==================== Check the results ================================
    print("comp\toutlet.tot_molfrac")
    for i in model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp:
        print(str(i)+"\t"+str(value( model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp[i] )))
    print()

    # NOTE: Changed all to mole fraction
    for i in model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp:
        print(str(i)+"\t"+str(value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp[i])))
    print()

    Ca = value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp["Liq","Ca_2+"])
    OH = value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp["Liq","OH_-"])
    H = value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp["Liq","H_+"])
    CO3 = value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp["Liq","CO3_2-"])
    Ksp_CaCO3 = value(model.fs.unit.control_volume.reactions[0.0].k_eq["CaCO3_Ksp"].expr)
    Ksp_CaOH2 = value(model.fs.unit.control_volume.reactions[0.0].k_eq["CaOH2_Ksp"].expr)

    print("Final pH = " +str(-log10(H*55.2)))

    print()
    print("Ksp_CaCO3 =\t"+str(Ksp_CaCO3))
    print("Ca*CO3 =\t"+str(Ca*CO3))
    print()
    if Ksp_CaCO3*1.01 >= Ca*CO3:
        print("Constraint is satisfied!")
    else:
        print("Constraint is VIOLATED!")
        print("\tRelative error: "+str(Ksp_CaCO3/Ca/CO3)+">=1")
        assert False

    print()
    print("Ksp_CaOH2 =\t"+str(Ksp_CaOH2))
    print("Ca*OH**2 =\t"+str(Ca*OH**2))
    print()
    if Ksp_CaOH2*1.01 >= Ca*OH**2:
        print("Constraint is satisfied!")
    else:
        print("Constraint is VIOLATED!")
        print("\tRelative error: "+str(Ksp_CaOH2/Ca/OH**2)+">=1")
        assert False

    print("==========================================================================")

    return model


@pytest.mark.component
def test_case_3_seawater_with_ca():
    model = run_case3(xOH=1e-7/55.2, xH=1e-7/55.2, xCaCO3=1e-20, xCaOH2 = 1e-20, xCa=200/50000/2/55.2,
                    xH2CO3=1e-20, xHCO3=0.00206/55.2, xCO3=1e-20,
                    thermo_config=case3_thermo_config, rxn_config=case3_log_rxn_config,
                    has_energy_balance=True)


@pytest.mark.component
def test_case_3_ultra_high_ca_forced_lime_precip():
    model = run_case3(xOH=1e-1/55.2, xH=1e-13/55.2, xCaCO3=1e-20, xCaOH2 = 1e-20, xCa=2000/50000/2/55.2,
                    xH2CO3=1e-20, xHCO3=0.00206/55.2, xCO3=1e-20,
                    thermo_config=case3_thermo_config, rxn_config=case3_log_rxn_config,
                    has_energy_balance=True)


@pytest.mark.component
def test_case_3_ultra_faux_added_lime_and_ash():
    added_lime_x = 100/50000/2/55.2
    added_ash_x = 2000/50000/2/55.2
    model = run_case3(xOH=(1e-7/55.2)+2*added_lime_x, xH=1e-7/55.2, xCaCO3=1e-20, xCaOH2 = 1e-20, xCa=(2000/50000/2/55.2)+added_lime_x,
                    xH2CO3=1e-20, xHCO3=0.00206/55.2, xCO3=1e-20+added_ash_x,
                    thermo_config=case3_thermo_config, rxn_config=case3_log_rxn_config,
                    has_energy_balance=True)


# Case 4 Config
'''
    Case 4: (phosphorus removal, most realistic test)
        Aqueous Rxns:
                        H2O <---> H + OH
                        H2CO3 <---> H + HCO3
                        HCO3 <---> H + CO3
                        H2PO4 <---> H + HPO4
                        HPO4 <---> H + PO4
                        FeOH <---> Fe + OH
                        Fe(OH)2 <---> FeOH + OH
                        Fe(OH)3 <---> Fe(OH)2 + OH
                        Fe(OH)4 <---> Fe(OH)3 + OH
        Solubility:
                        FePO4 <---> Fe + PO4
'''
case4_thermo_config = {
    "components": {
        'H2O': {"type": Solvent, "valid_phase_types": PT.aqueousPhase,
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (18.0153, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-285, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (69, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'H_+': {"type": Cation, "charge": 1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (1.00784, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (0, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (0, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'OH_-': {"type": Anion, "charge": -1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (17.008, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-230, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (-10, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'H2CO3': {"type": Solute, "valid_phase_types": PT.aqueousPhase,
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (62.03, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-699.7, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (187, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'HCO3_-': {"type": Anion, "charge": -1,
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (62.03, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-692, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (91.2, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'CO3_2-': {"type": Anion, "charge": -2,
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (62.03, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-677.1, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (-56.9, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'H2PO4_-': {"type": Anion, "charge": -1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (96.98724, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "enth_mol_form_liq_comp_ref": (-1296.3, pyunits.kJ/pyunits.mol),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                     "entr_mol_form_liq_comp_ref": (90.4, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'HPO4_2-': {"type": Anion, "charge": -2,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (95.9793, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "enth_mol_form_liq_comp_ref": (-1292.1, pyunits.kJ/pyunits.mol),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                     "entr_mol_form_liq_comp_ref": (-33.4, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'PO4_3-': {"type": Anion, "charge": -3,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (94.97136, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "enth_mol_form_liq_comp_ref": (-1277.4, pyunits.kJ/pyunits.mol),
                    "cp_mol_liq_comp_coeff": (75.312, pyunits.J/pyunits.mol/pyunits.K),
                     "entr_mol_form_liq_comp_ref": (-222, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'Fe_3+': {"type": Cation, "charge": 3,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (55.845, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (233467, pyunits.J/pyunits.kmol/pyunits.K),
                    "enth_mol_form_liq_comp_ref": (-48.5, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (-315.9, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'FeOH_2+': {"type": Cation, "charge": 2,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (72.8, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (305000, pyunits.J/pyunits.kmol/pyunits.K),
                    # NOTE: these parameters below are not well known
                    "enth_mol_form_liq_comp_ref": (-229.4, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (0, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'Fe(OH)2_+': {"type": Cation, "charge": 1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (89.8, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (375000, pyunits.J/pyunits.kmol/pyunits.K),
                    # NOTE: these parameters below are not well known
                    "enth_mol_form_liq_comp_ref": (-446.7, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (0, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'Fe(OH)3': {"type": Solute, "valid_phase_types": PT.aqueousPhase,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (106.8, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (446000, pyunits.J/pyunits.kmol/pyunits.K),
                    # NOTE: these parameters below are not well known
                    "enth_mol_form_liq_comp_ref": (-638.5, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (0, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'Fe(OH)4_-': {"type": Anion, "charge": -1,
              # Define the methods used to calculate the following properties
              "dens_mol_liq_comp": Constant,
              "enth_mol_liq_comp": Constant,
              "cp_mol_liq_comp": Constant,
              "entr_mol_liq_comp": Constant,
              # Parameter data is always associated with the methods defined above
              "parameter_data": {
                    "mw": (123.8, pyunits.g/pyunits.mol),
                    "dens_mol_liq_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_liq_comp_coeff": (518000, pyunits.J/pyunits.kmol/pyunits.K),
                    # NOTE: these parameters below are not well known
                    "enth_mol_form_liq_comp_ref": (-830.0, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_liq_comp_ref": (0, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    # End parameter_data
                    },
        'FePO4': {"type": Component, "valid_phase_types": PT.solidPhase,
              "dens_mol_sol_comp": Constant,
              "enth_mol_sol_comp": Constant,
              "cp_mol_sol_comp": Constant,
              "entr_mol_sol_comp": Constant,
              "parameter_data": {
                    "mw": (150.8, pyunits.g/pyunits.mol),
                    "dens_mol_sol_comp_coeff": (55.2, pyunits.kmol*pyunits.m**-3),
                    "cp_mol_sol_comp_coeff": (635000, pyunits.J/pyunits.kmol/pyunits.K),
                    "enth_mol_form_sol_comp_ref": (0, pyunits.kJ/pyunits.mol),
                    "entr_mol_form_sol_comp_ref": (0, pyunits.J/pyunits.K/pyunits.mol)
                                },
                    },

              },
              # End Component list
        "phases":  {'Liq': {"type": AqueousPhase,
                            "equation_of_state": Ideal},
                    'Sol': {"type": SolidPhase,
                            "equation_of_state": Ideal}
                    },

        "state_definition": FpcTP,
        "state_bounds": {
                         "temperature": (273.15, 300, 650),
                         "pressure": (5e4, 1e5, 1e6)
                     },

        "pressure_ref": 1e5,
        "temperature_ref": 300,
        "base_units": {"time": pyunits.s,
                       "length": pyunits.m,
                       "mass": pyunits.kg,
                       "amount": pyunits.mol,
                       "temperature": pyunits.K},

    }
    # End thermo_config definition

# Case 4 rxn config (with log_solubility_product)
case4_log_rxn_config = {
    "base_units": {"time": pyunits.s,
                   "length": pyunits.m,
                   "mass": pyunits.kg,
                   "amount": pyunits.mol,
                   "temperature": pyunits.K},
    "equilibrium_reactions": {
        "FePO4_Ksp": {
                    "stoichiometry": {  ("Sol", "FePO4"): -1,
                                        ("Liq", "Fe_3+"): 1,
                                        ("Liq", "PO4_3-"): 1},
                    "heat_of_reaction": constant_dh_rxn,
                    "equilibrium_constant": van_t_hoff,
                    "equilibrium_form": log_solubility_product,
                    "concentration_form": ConcentrationForm.moleFraction,
                    "parameter_data": {
                        "dh_rxn_ref": (0.0, pyunits.J/pyunits.mol),
                        "k_eq_ref": (10**-23/55.2/55.2, pyunits.dimensionless),
                        "T_eq_ref": (298.0, pyunits.K),
                        "reaction_order": { ("Sol", "FePO4"): 0,
                                            ("Liq", "Fe_3+"): 1,
                                            ("Liq", "PO4_3-"): 1}
                        }
                        # End parameter_data
                },
        "H2O_Kw": {
                "stoichiometry": {("Liq", "H2O"): -1,
                                 ("Liq", "H_+"): 1,
                                 ("Liq", "OH_-"): 1},
               "heat_of_reaction": constant_dh_rxn,
               "equilibrium_constant": van_t_hoff,
               "equilibrium_form": log_power_law_equil,
               "concentration_form": ConcentrationForm.moleFraction,
               "parameter_data": {
                   "dh_rxn_ref": (55.830, pyunits.J/pyunits.mol),
                   "k_eq_ref": (10**-14/55.2/55.2, pyunits.dimensionless),
                   "T_eq_ref": (298, pyunits.K),

                   # By default, reaction orders follow stoichiometry
                   #    manually set reaction order here to override
                   "reaction_order": {("Liq", "H2O"): 0,
                                    ("Liq", "H_+"): 1,
                                    ("Liq", "OH_-"): 1}
                    }
                    # End parameter_data
               },
        "H2CO3_Ka1": {
                "stoichiometry": {
                    ("Liq", "H2CO3"): -1,
                    ("Liq", "H_+"): 1,
                    ("Liq", "HCO3_-"): 1,
                },
                "heat_of_reaction": constant_dh_rxn,
                "equilibrium_constant": van_t_hoff,
                "equilibrium_form": log_power_law_equil,
                "concentration_form": ConcentrationForm.moleFraction,
                "parameter_data": {
                    "dh_rxn_ref": (7.7, pyunits.kJ/pyunits.mol),
                    "k_eq_ref": (10**-6.35/55.2, pyunits.dimensionless),
                    "T_eq_ref": (298, pyunits.K),
                    # By default, reaction orders follow stoichiometry
                    #    manually set reaction order here to override
                    "reaction_order": {
                        ("Liq", "H2CO3"): -1,
                        ("Liq", "H_+"): 1,
                        ("Liq", "HCO3_-"): 1,
                    },
                }
                # End parameter_data
            },
        "H2CO3_Ka2": {
                "stoichiometry": {
                    ("Liq", "HCO3_-"): -1,
                    ("Liq", "H_+"): 1,
                    ("Liq", "CO3_2-"): 1,
                },
                "heat_of_reaction": constant_dh_rxn,
                "equilibrium_constant": van_t_hoff,
                "equilibrium_form": log_power_law_equil,
                "concentration_form": ConcentrationForm.moleFraction,
                "parameter_data": {
                    "dh_rxn_ref": (14.9, pyunits.kJ/pyunits.mol),
                    "k_eq_ref": (10**-10.33/55.2, pyunits.dimensionless),
                    "T_eq_ref": (298, pyunits.K),
                    # By default, reaction orders follow stoichiometry
                    #    manually set reaction order here to override
                    "reaction_order": {
                        ("Liq", "HCO3_-"): -1,
                        ("Liq", "H_+"): 1,
                        ("Liq", "CO3_2-"): 1,
                    },
                }
                # End parameter_data
            },
        "H3PO4_Ka2": {
                    "stoichiometry": {  ("Liq", "H2PO4_-"): -1,
                                        ("Liq", "H_+"): 1,
                                        ("Liq", "HPO4_2-"): 1},
                    "heat_of_reaction": constant_dh_rxn,
                    "equilibrium_constant": van_t_hoff,
                    "equilibrium_form": log_power_law_equil,
                    "concentration_form": ConcentrationForm.moleFraction,
                    "parameter_data": {
                        "dh_rxn_ref": (4.2, pyunits.kJ/pyunits.mol),
                        "k_eq_ref": (10**-5.73/55.2, pyunits.dimensionless),
                        "T_eq_ref": (298, pyunits.K),

                        # By default, reaction orders follow stoichiometry
                        #    manually set reaction order here to override
                        "reaction_order": { ("Liq", "H2PO4_-"): -1,
                                            ("Liq", "H_+"): 1,
                                            ("Liq", "HPO4_2-"): 1}
                        }
                        # End parameter_data
                },
        "H3PO4_Ka3": {
                    "stoichiometry": {  ("Liq", "HPO4_2-"): -1,
                                        ("Liq", "H_+"): 1,
                                        ("Liq", "PO4_3-"): 1},
                    "heat_of_reaction": constant_dh_rxn,
                    "equilibrium_constant": van_t_hoff,
                    "equilibrium_form": log_power_law_equil,
                    "concentration_form": ConcentrationForm.moleFraction,
                    "parameter_data": {
                        "dh_rxn_ref": (14.7, pyunits.kJ/pyunits.mol),
                        "k_eq_ref": (10**-7.275/55.2, pyunits.dimensionless),
                        "T_eq_ref": (298, pyunits.K),

                        # By default, reaction orders follow stoichiometry
                        #    manually set reaction order here to override
                        "reaction_order": { ("Liq", "HPO4_2-"): -1,
                                            ("Liq", "H_+"): 1,
                                            ("Liq", "PO4_3-"): 1}
                        }
                        # End parameter_data
                },
        "FeOH_K": {
                    "stoichiometry": {  ("Liq", "FeOH_2+"): -1,
                                        ("Liq", "Fe_3+"): 1,
                                        ("Liq", "OH_-"): 1},
                    "heat_of_reaction": constant_dh_rxn,
                    "equilibrium_constant": van_t_hoff,
                    "equilibrium_form": log_power_law_equil,
                    "concentration_form": ConcentrationForm.moleFraction,
                    "parameter_data": {
                        "dh_rxn_ref": (0, pyunits.kJ/pyunits.mol),
                        "k_eq_ref": (1.768e-12/55.2, pyunits.dimensionless),
                        "T_eq_ref": (298, pyunits.K),

                        # By default, reaction orders follow stoichiometry
                        #    manually set reaction order here to override
                        "reaction_order": { ("Liq", "FeOH_2+"): -1,
                                            ("Liq", "Fe_3+"): 1,
                                            ("Liq", "OH_-"): 1}
                        }
                        # End parameter_data
                },
        "FeOH2_K": {
                    "stoichiometry": {  ("Liq", "Fe(OH)2_+"): -1,
                                        ("Liq", "FeOH_2+"): 1,
                                        ("Liq", "OH_-"): 1},
                    "heat_of_reaction": constant_dh_rxn,
                    "equilibrium_constant": van_t_hoff,
                    "equilibrium_form": log_power_law_equil,
                    "concentration_form": ConcentrationForm.moleFraction,
                    "parameter_data": {
                        "dh_rxn_ref": (0, pyunits.kJ/pyunits.mol),
                        "k_eq_ref": (3.757e-11/55.2, pyunits.dimensionless),
                        "T_eq_ref": (298, pyunits.K),

                        # By default, reaction orders follow stoichiometry
                        #    manually set reaction order here to override
                        "reaction_order": { ("Liq", "Fe(OH)2_+"): -1,
                                            ("Liq", "FeOH_2+"): 1,
                                            ("Liq", "OH_-"): 1}
                        }
                        # End parameter_data
                },
        "FeOH3_K": {
                    "stoichiometry": {  ("Liq", "Fe(OH)3"): -1,
                                        ("Liq", "Fe(OH)2_+"): 1,
                                        ("Liq", "OH_-"): 1},
                    "heat_of_reaction": constant_dh_rxn,
                    "equilibrium_constant": van_t_hoff,
                    "equilibrium_form": log_power_law_equil,
                    "concentration_form": ConcentrationForm.moleFraction,
                    "parameter_data": {
                        "dh_rxn_ref": (0, pyunits.kJ/pyunits.mol),
                        "k_eq_ref": (9.765e-7/55.2, pyunits.dimensionless),
                        "T_eq_ref": (298, pyunits.K),

                        # By default, reaction orders follow stoichiometry
                        #    manually set reaction order here to override
                        "reaction_order": { ("Liq", "Fe(OH)3"): -1,
                                            ("Liq", "Fe(OH)2_+"): 1,
                                            ("Liq", "OH_-"): 1}
                        }
                        # End parameter_data
                },
        "FeOH4_K": {
                    "stoichiometry": {  ("Liq", "Fe(OH)4_-"): -1,
                                        ("Liq", "Fe(OH)3"): 1,
                                        ("Liq", "OH_-"): 1},
                    "heat_of_reaction": constant_dh_rxn,
                    "equilibrium_constant": van_t_hoff,
                    "equilibrium_form": log_power_law_equil,
                    "concentration_form": ConcentrationForm.moleFraction,
                    "parameter_data": {
                        "dh_rxn_ref": (0, pyunits.kJ/pyunits.mol),
                        "k_eq_ref": (1.097e-6/55.2, pyunits.dimensionless),
                        "T_eq_ref": (298, pyunits.K),

                        # By default, reaction orders follow stoichiometry
                        #    manually set reaction order here to override
                        "reaction_order": { ("Liq", "Fe(OH)4_-"): -1,
                                            ("Liq", "Fe(OH)3"): 1,
                                            ("Liq", "OH_-"): 1}
                        }
                        # End parameter_data
                }
            #End last reaction
         }
         # End equilibrium_reactions
    }
    # End reaction_config definition

# Defaults to pH of 7
def run_case4(xOH=1e-7/55.2, xH=1e-7/55.2,
                xH2CO3=1e-20, xHCO3=1e-20, xCO3=1e-20,
                xH2PO4=1e-20, xHPO4=1e-20, xPO4=1e-20,
                xFe=1e-20, xFeOH=1e-20, xFeOH2=1e-20,
                xFeOH3=1e-20, xFeOH4=1e-20, xFePO4=1e-20,
                thermo_config=None, rxn_config=None,
                has_energy_balance=True, remove_precip_rxn=False):
    print("==========================================================================")
    print("Case 4: Phophorus removal through iron precipitation")
    print("xOH = "+str(xOH))
    print("xH = "+str(xH))
    print("xH2CO3 = "+str(xH2CO3))
    print("xHCO3 = "+str(xHCO3))
    print("xCO3 = "+str(xCO3))
    print("xH2PO4 = "+str(xH2PO4))
    print("xHPO4 = "+str(xHPO4))
    print("xPO4 = "+str(xPO4))

    print("xFe = "+str(xFe))
    print("xFeOH = "+str(xFeOH))
    print("xFeOH2 = "+str(xFeOH2))
    print("xFeOH3 = "+str(xFeOH3))
    print("xFeOH4 = "+str(xFeOH4))

    print("xFePO4 = "+str(xFePO4))
    print()

    Ksp_hold =  rxn_config["equilibrium_reactions"]["FePO4_Ksp"]["parameter_data"]["k_eq_ref"][0]
    if (remove_precip_rxn == True):
        del rxn_config["equilibrium_reactions"]["FePO4_Ksp"]

    model = ConcreteModel()
    model.fs = FlowsheetBlock(default={"dynamic": False})
    model.fs.thermo_params = GenericParameterBlock(default=thermo_config)

    model.fs.rxn_params = GenericReactionParameterBlock(
            default={"property_package": model.fs.thermo_params,
                    **rxn_config
                    })

    args = {"property_package": model.fs.thermo_params,
            "reaction_package": model.fs.rxn_params,
            "has_rate_reactions": False,
            "has_equilibrium_reactions": True,
            "has_heat_transfer": False,
            "has_heat_of_reaction": False,
            "has_pressure_change": False}
    if has_energy_balance == False:
        args["energy_balance_type"] = EnergyBalanceType.none

    model.fs.unit = EquilibriumReactor(default=args)

    total_flow_mol = 10

    # Set flow_mol_phase_comp
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H_+"].fix( xH*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "OH_-"].fix( xOH*total_flow_mol )

    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2CO3"].fix( xH2CO3*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "HCO3_-"].fix( xHCO3*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "CO3_2-"].fix( xCO3*total_flow_mol )

    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2PO4_-"].fix( xH2PO4*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "HPO4_2-"].fix( xHPO4*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "PO4_3-"].fix( xPO4*total_flow_mol )

    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Fe_3+"].fix( xFe*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "FeOH_2+"].fix( xFeOH*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Fe(OH)2_+"].fix( xFeOH2*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Fe(OH)3"].fix( xFeOH3*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Fe(OH)4_-"].fix( xFeOH4*total_flow_mol )

    model.fs.unit.inlet.flow_mol_phase_comp[0, "Sol", "FePO4"].fix( xFePO4*total_flow_mol )
    model.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(
        (1-xH-xOH-xFePO4-xH2CO3-xHCO3-xCO3-xH2PO4-xHPO4-xPO4-xFe-xFeOH-xFeOH2-xFeOH3-xFeOH4)*total_flow_mol )

    model.fs.unit.inlet.pressure.fix(101325.0)
    model.fs.unit.inlet.temperature.fix(298.)
    if has_energy_balance == False:
        model.fs.unit.outlet.temperature.fix(298.)

    assert (degrees_of_freedom(model) == 0)

    ## ==================== Start Scaling for this problem ===========================
    # # TODO: NOTE: The _set_eps_vals function may need to change. Some cases run
    #               better without it included.
    _set_eps_vals(model.fs.rxn_params, rxn_config)
    _set_equ_rxn_scaling(model.fs.unit, rxn_config)
    _set_mat_bal_scaling_FpcTP(model.fs.unit)
    if has_energy_balance == True:
        _set_ene_bal_scaling(model.fs.unit)

    iscale.calculate_scaling_factors(model.fs.unit)
    assert isinstance(model.fs.unit.control_volume.scaling_factor, Suffix)
    assert isinstance(model.fs.unit.control_volume.properties_out[0.0].scaling_factor, Suffix)
    assert isinstance(model.fs.unit.control_volume.properties_in[0.0].scaling_factor, Suffix)

    ## ==================== END Scaling for this problem ===========================

    model.fs.unit.initialize(optarg=solver.options, outlvl=idaeslog.DEBUG)

    assert degrees_of_freedom(model) == 0

    results = solver.solve(model, tee=True)

    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    ## ==================== Check the results ================================
    print("comp\toutlet.tot_molfrac")
    for i in model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp:
        print(str(i)+"\t"+str(value( model.fs.unit.control_volume.properties_out[0.0].mole_frac_comp[i] )))
    print()

    # NOTE: Changed all to mole fraction
    for i in model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp:
        print(str(i)+"\t"+str(value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp[i])))
    print()

    Fe = value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp["Liq","Fe_3+"])
    OH = value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp["Liq","OH_-"])
    H = value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp["Liq","H_+"])
    PO4 = value(model.fs.unit.control_volume.properties_out[0.0].mole_frac_phase_comp["Liq","PO4_3-"])
    if (remove_precip_rxn == False):
        Ksp = value(model.fs.unit.control_volume.reactions[0.0].k_eq["FePO4_Ksp"].expr)
    else:
        Ksp = Ksp_hold

    print("Final pH = " +str(-log10(H*55.2)))

    print()
    print("Ksp =\t"+str(Ksp))
    print("Fe*PO4 =\t"+str(Fe*PO4))
    print()
    if Ksp*1.01 >= Fe*PO4:
        print("Constraint is satisfied!")
    else:
        print("Constraint is VIOLATED!")
        print("\tRelative error: "+str(Ksp/Fe/PO4)+">=1")
        if (remove_precip_rxn == False):
            assert False

    print("==========================================================================")

    return model

# seawater with standard amount of iron (5.38e-8 M) and phosphorus (3.22e-6 M)
#   Boost Fe, PO4, and OH to induce precipitation
#       Add 1e-4 M of each...
@pytest.mark.component
def test_case_4_seawater_added_Fe_OH_PO4():
    model = run_case4(xOH=(1e-7/55.2)+(1e-4/55.2), xH=1e-7/55.2,
                    xH2CO3=1e-20, xHCO3=0.00206/55.2, xCO3=1e-20,
                    xH2PO4=1e-20, xHPO4=(3.22e-6/55.2)+(1e-4/55.2), xPO4=1e-20,
                    xFe=(5.38e-8/55.2)+(1e-4/55.2), xFeOH=1e-20, xFeOH2=1e-20,
                    xFeOH3=1e-20, xFeOH4=1e-20, xFePO4=1e-20,
                    thermo_config=case4_thermo_config,
                    rxn_config=case4_log_rxn_config,
                    has_energy_balance=True, remove_precip_rxn=False)

# seawater with standard amount of iron (5.38e-8 M) and phosphorus (3.22e-6 M)
#   Boost Fe, PO4, and OH to induce precipitation
#       Add 1e-4 M of Fe
#       Add 1e-4 M of PO4
#       Add 0 M of OH
@pytest.mark.component
def test_case_4_seawater_added_Fe_PO4():
    model = run_case4(xOH=(1e-7/55.2)+(0/55.2), xH=1e-7/55.2,
                    xH2CO3=1e-20, xHCO3=0.00206/55.2, xCO3=1e-20,
                    xH2PO4=1e-20, xHPO4=(3.22e-6/55.2)+(1e-4/55.2), xPO4=1e-20,
                    xFe=(5.38e-8/55.2)+(1e-4/55.2), xFeOH=1e-20, xFeOH2=1e-20,
                    xFeOH3=1e-20, xFeOH4=1e-20, xFePO4=1e-20,
                    thermo_config=case4_thermo_config,
                    rxn_config=case4_log_rxn_config,
                    has_energy_balance=True, remove_precip_rxn=False)

# seawater with standard amount of iron (5.38e-8 M) and phosphorus (3.22e-6 M)
#   Boost Fe, PO4, and OH to induce precipitation
#       Add 1e-4 M of Fe
#       Add 0 M of PO4
#       Add 0 M of OH
@pytest.mark.component
def test_case_4_seawater_added_Fe_only():
    model = run_case4(xOH=(1e-7/55.2)+(0/55.2), xH=1e-7/55.2,
                    xH2CO3=1e-20, xHCO3=0.00206/55.2, xCO3=1e-20,
                    xH2PO4=1e-20, xHPO4=(3.22e-6/55.2)+(0/55.2), xPO4=1e-20,
                    xFe=(5.38e-8/55.2)+(1e-4/55.2), xFeOH=1e-20, xFeOH2=1e-20,
                    xFeOH3=1e-20, xFeOH4=1e-20, xFePO4=1e-20,
                    thermo_config=case4_thermo_config,
                    rxn_config=case4_log_rxn_config,
                    has_energy_balance=True, remove_precip_rxn=False)

# This is for additional testing
if __name__ == "__main__":
    # seawater with standard amount of iron (5.38e-8 M) and phosphorus (3.22e-6 M)
    #
    #       This works, but convergence of the initialization stage is VERY, VERY poor!!!
    #
    #   # TODO:  Figure out how to better initialize this case.
    # # TODO: NOTE: The _set_eps_vals function may need to change. Some cases run
    #               better without it included.

    model = run_case4(xOH=1e-7/55.2, xH=1e-7/55.2,
                    xH2CO3=1e-20, xHCO3=0.00206/55.2, xCO3=1e-20,
                    xH2PO4=1e-20, xHPO4=3.22e-6/55.2, xPO4=1e-20,
                    xFe=5.38e-8/55.2, xFeOH=1e-20, xFeOH2=1e-20,
                    xFeOH3=1e-20, xFeOH4=1e-20, xFePO4=1e-20,
                    thermo_config=case4_thermo_config,
                    rxn_config=case4_log_rxn_config,
                    has_energy_balance=True, remove_precip_rxn=False)

    exit()



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
import pytest
from pyomo.environ import ConcreteModel
from idaes.core import FlowsheetBlock
from watertap.examples.flowsheets.full_treatment_train.flowsheet_components import pretreatment_NF
from watertap.examples.flowsheets.full_treatment_train.model_components import property_models
from watertap.examples.flowsheets.full_treatment_train.util import check_build, check_scaling


@pytest.mark.unit
def test_build_and_scale_pretreatment_NF():
    for has_bypass in [True, False]:
        for NF_type in ['Sep', 'ZO']:
            for NF_base in ['ion', 'salt']:
                kwargs = {'has_bypass': has_bypass, 'NF_type': NF_type, 'NF_base': NF_base}
                print('\n\n***kwargs****\n', kwargs)

                if NF_type == 'ZO' and NF_base == 'salt':
                    continue  # not a valid combination

                m = ConcreteModel()
                m.fs = FlowsheetBlock(default={"dynamic": False})
                property_models.build_prop(m, base=kwargs['NF_base'])

                check_build(m, build_func=pretreatment_NF.build_pretreatment_NF, **kwargs)
                assert hasattr(m.fs, 'NF')
                check_scaling(m, scale_func=pretreatment_NF.scale_pretreatment_NF, **kwargs)

                pretreatment_NF.display_pretreatment_NF(m, **kwargs)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

from pyomo.environ import Constraint
# Import IDAES cores
from idaes.generic_models.unit_models.pressure_changer import PumpData
from idaes.core import declare_process_block_class
import idaes.core.util.scaling as iscale

import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)

@declare_process_block_class("Pump")
class PumpIsothermalData(PumpData):
    """
    Standard Isothermal Pump Unit Model Class
    """

    def build(self):
        super().build()

        self.control_volume.del_component(self.control_volume.enthalpy_balances)

        @self.control_volume.Constraint(
            self.flowsheet().config.time,
            doc="Isothermal constraint")
        def isothermal_balance(b, t):
            return b.properties_in[t].temperature == b.properties_out[t].temperature

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        for ind, c in self.control_volume.isothermal_balance.items():
            sf = iscale.get_scaling_factor(self.control_volume.properties_in[0].temperature)
            iscale.constraint_scaling_transform(c, sf)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

"""
    These are a set of utils for aiding in solving flowsheets with chemical
    reactions. Based on initiale testing, it was found that these tools may
    be necessary to get a flowsheet that includes electrolyte chemistry to
    converge properly.
"""

from pyomo.network import SequentialDecomposition
from idaes.core.util import scaling as iscale
from idaes.core.util import get_solver
from pyomo.environ import value

__author__ = "Austin Ladshaw"

# Get default solver for testing
solver = get_solver()

def set_H2O_molefraction(port):
    # Perform a summation of all non-H2O molefractions to find the H2O molefraction
    sum = 0
    for i in port.mole_frac_comp:
        # NOTE: i will be a tuple with format (time, component)
        if i[1] != "H2O":
            sum += value(port.mole_frac_comp[i[0], i[1]])

    port.mole_frac_comp[0, "H2O"].set_value( 1-sum )

def zero_out_non_H2O_molefractions(port):
    for i in port.mole_frac_comp:
        if i[1] != "H2O":
            port.mole_frac_comp[i[0], i[1]].set_value(0)

def fix_all_molefractions(port):
    for i in port.mole_frac_comp:
        port.mole_frac_comp[i[0], i[1]].fix()

def unfix_all_molefractions(port):
    for i in port.mole_frac_comp:
        port.mole_frac_comp[i[0], i[1]].unfix()

# NOTE: With new log conc form, we don't seem to need the very low
#       bound push anymore. In fact, too low a value is degrading
#       the solutions/iterations in some cases.
#
#   This tool for block decomp may be less useful moving forward.
#       probably better to initialize each manually.
def block_initializer(blk, tee=False):
    solver.options['bound_push'] = 1e-1
    solver.options['mu_init'] = 1e-1
    solver.options["nlp_scaling_method"] = "user-scaling"
    if tee==True:
        print(blk)
    results = solver.solve(blk, tee=tee)

    # This is a temporary fix to get around some errors in RO with detailed level
    try:
        iscale.constraint_autoscale_large_jac(blk)
    except:
        pass

def seq_decomp_initializer(model):
    seq = SequentialDecomposition(tol=1.0E-3)
    seq.options.select_tear_method = "heuristic"
    seq.run(model, block_initializer)

    # This is a temporary fix to get around some errors in RO with detailed level
    try:
        iscale.constraint_autoscale_large_jac(model)
    except:
        pass



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
'''
mutable parameters for optimization:
    m.fs.system_recovery_target
    m.fs.max_allowable_pressure
'''
from pyomo.environ import ConcreteModel, TransformationFactory, Param, value
from pyomo.environ import units as pyunits

from idaes.core import FlowsheetBlock
from idaes.core.util.initialization import propagate_state
from idaes.core.util.scaling import calculate_scaling_factors

from watertap.examples.flowsheets.full_treatment_train.util import (solve_with_user_scaling,
                                                             check_dof)

import watertap.examples.flowsheets.full_treatment_train.analysis.flowsheet_single_stage as single_stage
from watertap.examples.flowsheets.full_treatment_train.analysis.flowsheet_single_stage import (
        build, build_components, scale, initialize, report, optimize)


desal_kwargs = {'has_desal_feed': False, 'is_twostage': True, 'has_ERD': True,
        'RO_type': '0D', 'RO_base': 'TDS', 'RO_level': 'detailed'}


def set_optimization_components(m, system_recovery, **kwargs):
    single_stage.set_optimization_components(m, system_recovery, **kwargs)

    m.fs.max_allowable_pressure = Param(initialize=300e5, mutable=True, units=pyunits.pascal)
    m.fs.pump_RO2.control_volume.properties_out[0].pressure.unfix()
    m.fs.pump_RO2.control_volume.properties_out[0].pressure.setlb(20e5)
    m.fs.pump_RO2.control_volume.properties_out[0].pressure.setub(m.fs.max_allowable_pressure)

    m.fs.RO2.area.unfix()
    m.fs.RO2.area.setlb(1)
    m.fs.RO2.area.setub(300)

    if kwargs['RO_type'] == '0D':
        m.fs.RO2.N_Re_io[0, 'in'].unfix()
    elif kwargs['RO_type'] == '1D':
        m.fs.RO2.N_Re[0, 0].unfix()

    # Set lower bound for water flux at the RO outlet, based on a minimum net driving pressure, NDPmin
    m.fs.RO2.NDPmin = Param(initialize=1e5, mutable=True, units=pyunits.Pa)
    if kwargs['RO_type'] == '0D':
        m.fs.RO2.flux_mass_io_phase_comp[0, 'out', 'Liq', 'H2O'].setlb(value(m.fs.RO2.A_comp[0, 'H2O']
                                                                             * m.fs.RO2.dens_solvent
                                                                             * m.fs.RO2.NDPmin))
    elif kwargs['RO_type'] == '1D':
        m.fs.RO2.flux_mass_phase_comp[0, 1, 'Liq', 'H2O'].setlb(value(m.fs.RO2.A_comp[0, 'H2O']
                                                                      * m.fs.RO2.dens_solvent
                                                                      * m.fs.RO2.NDPmin))


def set_up_optimization(m, system_recovery=0.50, **kwargs):
    set_optimization_components(m, system_recovery, **kwargs)
    check_dof(m, 6)


def solve_flowsheet(**desal_kwargs):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    build(m, **desal_kwargs)
    TransformationFactory("network.expand_arcs").apply_to(m)

    # scale
    scale(m, **desal_kwargs)
    calculate_scaling_factors(m)

    # initialize
    m.fs.feed.initialize()
    propagate_state(m.fs.s_pretrt_tb)
    initialize(m, **desal_kwargs)

    check_dof(m)
    solve_with_user_scaling(m, tee=False, fail_flag=True)

    # report
    print('==================================='
          '\n          Simulation          ')
    report(m, **desal_kwargs)

    return m


def optimize_flowsheet(system_recovery=0.50, **kwargs):
    m = solve_flowsheet(**kwargs)
    set_up_optimization(m, system_recovery=system_recovery, **kwargs)
    optimize(m)
    print('==================================='
          '\n       Optimization            ')
    report(m, **kwargs)

    return m


if __name__ == "__main__":
    import sys
    if len(sys.argv) == 1:
        m = solve_flowsheet(**desal_kwargs)
    else:
        m = optimize_flowsheet(system_recovery=float(sys.argv[1]), **desal_kwargs)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

import pyomo.environ as pyo
from pyomo.core.base.block import _BlockData
from pyomo.core.kernel.block import IBlock
from pyomo.solvers.plugins.solvers.IPOPT import IPOPT

from idaes.core.util.scaling import (constraint_autoscale_large_jac,
        get_scaling_factor, set_scaling_factor, unset_scaling_factor)
from idaes.logger import getLogger

_log = getLogger("watertap.core")


@pyo.SolverFactory.register("ipopt-watertap",
        doc="The Ipopt NLP solver, with user-based variable and automatic Jacobian constraint scaling")
class IpoptWaterTAP(IPOPT):

    def __init__(self, **kwds):
        kwds["name"] = "ipopt-watertap"
        super().__init__(**kwds)

    def _presolve(self, *args, **kwds):
        if len(args) > 1 or len(args) == 0:
            raise TypeError(f"IpoptWaterTAP.solve takes 1 positional argument but {len(args)} were given")
        if not isinstance(args[0], (_BlockData, IBlock)):
            raise TypeError("IpoptWaterTAP.solve takes 1 positional argument: a Pyomo ConcreteModel or Block")

        self._tee = kwds.get("tee", False)

        # Set the IDAES-default tol
        if "tol" not in self.options:
            self.options["tol"] = 1e-6

        if not self._is_user_scaling():
            self._reset_needed = False
            super()._presolve(*args, **kwds)
            return

        if self._tee:
            print("ipopt-watertap: Ipopt with user variable scaling and IDAES jacobian constraint scaling")

        # These options are typically available with gradient-scaling, and they
        # have corresponding options in the IDAES constraint_autoscale_large_jac
        # function. Here we use their Ipopt names and default values, see
        # https://coin-or.github.io/Ipopt/OPTIONS.html#OPT_NLP_Scaling
        max_grad = self._get_option("nlp_scaling_max_gradient", 100)
        min_scale = self._get_option("nlp_scaling_min_value", 1e-8)

        # These options are custom for the IDAES constraint_autoscale_large_jac
        # function. We expose them as solver options as this has become part
        # of the solve process.
        ignore_variable_scaling = self._get_option("ignore_variable_scaling", False)
        ignore_constraint_scaling = self._get_option("ignore_constraint_scaling", False)

        self._model = args[0]
        self._cache_scaling_factors()
        self._reset_needed = True

        # NOTE: This function sets the scaling factors on the
        #       constraints. Hence we cache the constraint scaling
        #       factors and reset them to their original values
        #       so that repeated calls to solve change the scaling
        #       each time based on the initial values, just like in Ipopt.
        try:
            constraint_autoscale_large_jac(self._model,
                    ignore_constraint_scaling=ignore_constraint_scaling,
                    ignore_variable_scaling=ignore_variable_scaling,
                    max_grad=max_grad,
                    min_scale=min_scale)
        except AssertionError as err:
            if str(err) == "Error in AMPL evaluation":
                print("ipopt-watertap: Issue in AMPL function evaluation; Jacobian constraint scaling not applied.")
                halt_on_ampl_error = self.options.get("halt_on_ampl_error", "yes")
                if halt_on_ampl_error == "no" :
                    print("ipopt-watertap: halt_on_ampl_error=no, so continuing with optimization.")
                else:
                    raise RuntimeError("Error in AMPL evaluation.\n"
                            "Run ipopt with halt_on_ampl_error=yes and symbolic_solver_labels=True to see the affected function.")
            else:
                print("Error in constraint_autoscale_large_jac")
                raise

        # this creates the NL file, among other things
        super()._presolve(*args, **kwds)

    def _postsolve(self):
        if self._reset_needed:
            self._reset_scaling_factors()
            # remove our reference to the model
            del self._model
        del self._reset_needed
        return super()._postsolve()

    def _cache_scaling_factors(self):
        self._scaling_cache = [ (c, get_scaling_factor(c)) for c in
                self._model.component_data_objects(
                    pyo.Constraint, active=True, descend_into=True)]

    def _reset_scaling_factors(self):
        for c, s in self._scaling_cache:
            if s is None:
                unset_scaling_factor(c)
            else:
                set_scaling_factor(c, s)
        del self._scaling_cache

    def _get_option(self, option_name, default_value):
        # NOTE: options get reset to their original value at the end of the
        #       OptSolver.solve. The options in _presolve (where this is called)
        #       are already copies of the original, so it is safe to pop them so
        #       they don't get sent to Ipopt.
        option_value = self.options.pop(option_name, None)
        if option_value is None:
            option_value = default_value
        else:
            if self._tee:
                print(f"ipopt-watertap: {option_name}={option_value}")
        return option_value

    def _is_user_scaling(self):
        if "nlp_scaling_method" not in self.options:
            self.options["nlp_scaling_method"] = "user-scaling"
        if self.options["nlp_scaling_method"] != "user-scaling":
            if self._tee:
                print("The ipopt-watertap solver is designed to be run with user-scaling. "
                        f"Ipopt with nlp_scaling_method={self.options['nlp_scaling_method']} will be used instead")
            return False
        return True


## reconfigure IDAES to use the ipopt-watertap solver
import idaes
_default_solver_config_value = idaes.cfg.get("default_solver")
_idaes_default_solver = _default_solver_config_value._default

_default_solver_config_value.set_default_value("ipopt-watertap")
if not _default_solver_config_value._userSet:
    _default_solver_config_value.reset()



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

"""Pretreatment flowsheet components"""

from pyomo.environ import ConcreteModel, TransformationFactory
from pyomo.network import Arc
from idaes.core import FlowsheetBlock
from watertap.examples.flowsheets.full_treatment_train.model_components import Separator, Mixer
from idaes.generic_models.unit_models.separator import SplittingType, EnergySplittingType
from idaes.core.util.scaling import (calculate_scaling_factors,
                                     set_scaling_factor,
                                     get_scaling_factor,
                                     constraint_scaling_transform)
from idaes.core.util.initialization import propagate_state
from watertap.unit_models.pump_isothermal import Pump
from watertap.examples.flowsheets.full_treatment_train.flowsheet_components import feed_block
from watertap.examples.flowsheets.full_treatment_train.model_components import unit_separator, unit_ZONF, property_models
from watertap.examples.flowsheets.full_treatment_train.util import solve_with_user_scaling, check_dof


def build_pretreatment_NF(m, has_bypass=True, NF_type='ZO', NF_base='ion'):
    """
    Builds NF pretreatment including specified feed and auxiliary equipment.
    Arguments:
        has_bypass: True or False, default = True
        NF_type: 'Sep' or 'ZO', default = 'ZO'
        NF_base: 'ion' or 'salt', default = 'ion'
    """
    pretrt_port = {}
    prop = property_models.get_prop(m, base=NF_base)

    # build feed
    feed_block.build_feed(m, base=NF_base)

    # build NF
    if NF_type == 'Sep':
        unit_separator.build_SepNF(m, base=NF_base)
    elif NF_type == 'ZO':
        unit_ZONF.build_ZONF(m, base=NF_base)
        m.fs.pump_NF = Pump(default={'property_package': prop})
    else:
        raise ValueError('Unexpected model type {NF_type} provided to build_NF_no_bypass'
                         ''.format(NF_type=NF_type))

    if has_bypass:
        # build auxiliary units
        m.fs.splitter = Separator(default={
            "property_package": prop,
            "outlet_list": ['pretreatment', 'bypass'],
            "split_basis": SplittingType.totalFlow,
            "energy_split_basis": EnergySplittingType.equal_temperature})
        m.fs.mixer = Mixer(default={
            "property_package": prop,
            "inlet_list": ['pretreatment', 'bypass']})

        # connect models
        m.fs.s_pretrt_feed_splitter = Arc(source=m.fs.feed.outlet, destination=m.fs.splitter.inlet)
        m.fs.s_pretrt_splitter_mixer = Arc(source=m.fs.splitter.bypass, destination=m.fs.mixer.bypass)
        if NF_type == 'ZO':
            m.fs.s_pretrt_splitter_pumpNF = Arc(source=m.fs.splitter.pretreatment, destination=m.fs.pump_NF.inlet)
            m.fs.s_pretrt_pumpNF_NF = Arc(source=m.fs.pump_NF.outlet, destination=m.fs.NF.inlet)
        else:
            m.fs.s_pretrt_splitter_NF = Arc(source=m.fs.splitter.pretreatment, destination=m.fs.NF.inlet)
        m.fs.s_pretrt_NF_mixer = Arc(source=m.fs.NF.permeate, destination=m.fs.mixer.pretreatment)

        # specify (NF and feed is already specified, mixer has 0 DOF, splitter has 1 DOF, NF pump has 2 DOF)
        # splitter
        m.fs.splitter.split_fraction[0, 'bypass'].fix(0.1)
        if NF_type == 'ZO':
            m.fs.pump_NF.efficiency_pump.fix(0.80)
            m.fs.pump_NF.control_volume.properties_out[0].pressure.fix(4e5)

        # inlet/outlet ports for pretreatment
        pretrt_port['out'] = m.fs.mixer.outlet
        pretrt_port['waste'] = m.fs.NF.retentate

    else:  # no bypass
        # build auxiliary units (none)

        # connect models
        if NF_type == 'ZO':
            m.fs.s_pretrt_feed_pumpNF = Arc(source=m.fs.feed.outlet, destination=m.fs.pump_NF.inlet)
            m.fs.s_pretrt_pumpNF_NF = Arc(source=m.fs.pump_NF.outlet, destination=m.fs.NF.inlet)
            #TODO: should source be m.fs.pump_NF.outlet? Double-check here and other arcs with pump_NF
        else:
            m.fs.s_pretrt_feed_NF = Arc(source=m.fs.feed.outlet, destination=m.fs.NF.inlet)

        # specify (NF and feed are already specified, NF pump has 2 DOF)
        if NF_type == 'ZO':
            m.fs.pump_NF.efficiency_pump.fix(0.80)
            m.fs.pump_NF.control_volume.properties_out[0].pressure.fix(4e5)

        # inlet/outlet ports for pretreatment
        pretrt_port['out'] = m.fs.NF.permeate
        pretrt_port['waste'] = m.fs.NF.retentate

    return pretrt_port


def scale_pretreatment_NF(m, **kwargs):
    calculate_scaling_factors(m.fs.feed)
    calculate_scaling_factors(m.fs.NF)

    if kwargs['has_bypass']:
        calculate_scaling_factors(m.fs.splitter)
        set_scaling_factor(m.fs.splitter.split_fraction, 1)  # TODO: should have an IDAES default
        constraint_scaling_transform(m.fs.splitter.sum_split_frac[0], 1)  # TODO: should have an IDAES default
        calculate_scaling_factors(m.fs.mixer)
        set_scaling_factor(m.fs.mixer.minimum_pressure,
                           get_scaling_factor(m.fs.mixer.mixed_state[0].pressure)
                           )  # TODO: IDAES should have a default and link to the constraint
        for c in [m.fs.mixer.minimum_pressure_constraint[0, 1],
                  m.fs.mixer.minimum_pressure_constraint[0, 2],
                  m.fs.mixer.mixture_pressure[0.0]]:
            constraint_scaling_transform(c, get_scaling_factor(m.fs.mixer.minimum_pressure))

    if kwargs['NF_type'] == 'ZO':
        set_scaling_factor(m.fs.pump_NF.control_volume.work, 1e-3)
        calculate_scaling_factors(m.fs.pump_NF)
        set_scaling_factor(m.fs.pump_NF.ratioP, 1)  # TODO: IDAES should have a default and link to the constraint

def initialize_pretreatment_NF(m, **kwargs):
    optarg = {'nlp_scaling_method': 'user-scaling'}

    if kwargs['has_bypass']:
        m.fs.feed.initialize(optarg=optarg)
        propagate_state(m.fs.s_pretrt_feed_splitter)
        m.fs.splitter.initialize(optarg=optarg)
        propagate_state(m.fs.s_pretrt_splitter_mixer)
        if kwargs['NF_type'] == 'ZO':
            propagate_state(m.fs.s_pretrt_splitter_pumpNF)
            m.fs.pump_NF.initialize(optarg=optarg)
            propagate_state(m.fs.s_pretrt_pumpNF_NF)
            m.fs.NF.initialize(optarg=optarg)
        else:  # NF_type == 'Sep'
            propagate_state(m.fs.s_pretrt_splitter_NF)
            # m.fs.NF.initialize(optarg=optarg)  # IDAES error when NF is a separator TODO: address in IDAES
        propagate_state(m.fs.s_pretrt_NF_mixer)
        m.fs.mixer.initialize(optarg=optarg)

    else:  # no bypass
        m.fs.feed.initialize(optarg=optarg)
        if kwargs['NF_type'] == 'ZO':
            propagate_state(m.fs.s_pretrt_feed_pumpNF)
            m.fs.pump_NF.initialize(optarg=optarg)
            propagate_state(m.fs.s_pretrt_pumpNF_NF)
            m.fs.NF.initialize(optarg=optarg)
        else:  # NF_type == 'Sep'
            propagate_state(m.fs.s_pretrt_feed_NF)
            # m.fs.NF.initialize(optarg=optarg)  # IDAES error when NF is a separator TODO: address in IDAES


def display_pretreatment_NF(m, **kwargs):

    m.fs.feed.report()

    if kwargs['has_bypass']:
        m.fs.splitter.report()
        if kwargs['NF_type'] == 'ZO':
            m.fs.pump_NF.report()
        m.fs.NF.report()
        m.fs.mixer.report()
    else:  # no bypass
        if kwargs['NF_type'] == 'ZO':
            m.fs.pump_NF.report()
        m.fs.NF.report()


def solve_pretreatment_NF(**kwargs):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    property_models.build_prop(m, base=kwargs['NF_base'])
    build_pretreatment_NF(m, **kwargs)
    TransformationFactory("network.expand_arcs").apply_to(m)

    scale_pretreatment_NF(m, **kwargs)

    initialize_pretreatment_NF(m, **kwargs)

    check_dof(m)
    solve_with_user_scaling(m, tee=True, fail_flag=True)

    display_pretreatment_NF(m, **kwargs)


if __name__ == "__main__":
    solve_pretreatment_NF(has_bypass=True, NF_type='ZO', NF_base='ion')



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

"""Zero order nanofiltration model based on specifying solvent flux and solute rejection"""

from pyomo.environ import ConcreteModel, Constraint
from idaes.core import FlowsheetBlock
from watertap.unit_models.nanofiltration_ZO import NanofiltrationZO
from idaes.core.util.scaling import calculate_scaling_factors, constraint_scaling_transform
from watertap.examples.flowsheets.full_treatment_train.model_components import property_models
from watertap.examples.flowsheets.full_treatment_train.util import solve_with_user_scaling, check_dof


def build_ZONF(m, base='ion'):
    """
    Builds a ZONF model based on specified rejection and solvent flux.
    Requires prop_ion property package.
    """

    if base not in ['ion']:
        raise ValueError('Unexpected property base {base} for build_ZONF'
                         ''.format(base=base))
    prop = property_models.get_prop(m, base=base)

    m.fs.NF = NanofiltrationZO(default={
        "property_package": prop,
        "has_pressure_change": False})

    # specify
    m.fs.NF.flux_vol_solvent.fix(1.67e-6)
    m.fs.NF.area.fix(500)
    m.fs.NF.properties_permeate[0].pressure.fix(101325)

    m.fs.NF.rejection_phase_comp[0, 'Liq', 'Na'].fix(0.01)
    m.fs.NF.rejection_phase_comp[0, 'Liq', 'Ca'].fix(0.79)
    m.fs.NF.rejection_phase_comp[0, 'Liq', 'Mg'].fix(0.94)
    m.fs.NF.rejection_phase_comp[0, 'Liq', 'SO4'].fix(0.87)
    m.fs.NF.rejection_phase_comp[0, 'Liq', 'Cl'] = 0.15  # guess, but electroneutrality enforced below
    charge_comp = {'Na': 1, 'Ca': 2, 'Mg': 2, 'SO4': -2, 'Cl': -1}
    m.fs.NF.eq_electroneutrality = Constraint(
        expr=0 ==
             sum(charge_comp[j]
                 * m.fs.NF.feed_side.properties_out[0].conc_mol_phase_comp['Liq', j]
                 for j in charge_comp))
    constraint_scaling_transform(m.fs.NF.eq_electroneutrality, 1)


def solve_ZONF(base='ion'):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    property_models.build_prop(m, base=base)
    build_ZONF(m, base=base)
    property_models.specify_feed(m.fs.NF.feed_side.properties_in[0], base='ion')

    check_dof(m)
    calculate_scaling_factors(m)
    solve_with_user_scaling(m)

    m.fs.NF.inlet.display()
    m.fs.NF.permeate.display()
    m.fs.NF.retentate.display()

    return m


if __name__ == "__main__":
    solve_ZONF(base='ion')




###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

"""
    Simple unit tests for example flowsheet of SepRO with Chlorination.

    NOTE: That flowsheet is not meant to be viewed as a final product, but
    a sample of how to incorporate more complex chemistry into a simple
    flowsheet.
"""
import pytest

from watertap.examples.flowsheets.full_treatment_train.flowsheet_components.chemistry.posttreatment_ideal_naocl_chlorination_block import (
    run_ideal_naocl_mixer_example, run_ideal_naocl_chlorination_example, run_chlorination_block_example,
    build_translator_from_RO_to_chlorination_block)
from watertap.examples.flowsheets.full_treatment_train.model_components import property_models

__author__ = "Austin Ladshaw"

@pytest.mark.component
def test_ideal_naocl_mixer():
    model = run_ideal_naocl_mixer_example()
    assert model.fs.ideal_naocl_mixer_unit.dosing_rate.value == \
            pytest.approx(0.4, rel=1e-3)
    assert model.fs.ideal_naocl_mixer_unit.outlet.flow_mol[0].value == \
            pytest.approx(10.00001074690966, rel=1e-3)
    assert model.fs.ideal_naocl_mixer_unit.outlet.mole_frac_comp[0, "OCl_-"].value == \
            pytest.approx(5.373448801531194e-07, rel=1e-3)

@pytest.mark.component
def test_ideal_naocl_chlorination():
    model = run_ideal_naocl_chlorination_example()
    assert model.fs.ideal_naocl_chlorination_unit.free_chlorine.value == \
            pytest.approx(1.893849296278168, rel=1e-3)
    assert model.fs.ideal_naocl_chlorination_unit.outlet.mole_frac_comp[0,'OCl_-'].value == \
            pytest.approx(4.254858076511e-07, rel=1e-3)
    assert model.fs.ideal_naocl_chlorination_unit.outlet.mole_frac_comp[0,'H_+'].value == \
            pytest.approx(5.6407676871845223e-11, rel=1e-3)

@pytest.mark.component
def test_ideal_naocl_chlorination_full_block():
    model = run_chlorination_block_example(fix_free_chlorine=True)
    assert model.fs.ideal_naocl_mixer_unit.dosing_rate.value == \
            pytest.approx(0.9504457542440085, rel=1e-3)
    assert model.fs.ideal_naocl_mixer_unit.outlet.flow_mol[0].value == \
            pytest.approx(25.000025535888078, rel=1e-3)
    assert model.fs.ideal_naocl_mixer_unit.outlet.mole_frac_comp[0, "OCl_-"].value == \
            pytest.approx(5.107172398859054e-07, rel=1e-3)

    assert model.fs.ideal_naocl_chlorination_unit.free_chlorine.value == \
            pytest.approx(2, rel=1e-3)
    assert model.fs.ideal_naocl_chlorination_unit.outlet.mole_frac_comp[0,'OCl_-'].value == \
            pytest.approx(4.508812116261189e-07, rel=1e-3)
    assert model.fs.ideal_naocl_chlorination_unit.outlet.mole_frac_comp[0,'H_+'].value == \
            pytest.approx(5.47976004923939e-11, rel=1e-3)

@pytest.mark.component
def test_addition_of_translator():
    model = run_ideal_naocl_mixer_example()
    property_models.build_prop(model, base='TDS')
    build_translator_from_RO_to_chlorination_block(model)
    assert hasattr(model.fs, 'RO_to_Chlor')



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

"""Seawater feed specifications for supported property packages"""

from pyomo.environ import ConcreteModel
from idaes.core import FlowsheetBlock
from idaes.generic_models.properties.core.generic.generic_property import GenericParameterBlock
from idaes.core.util.scaling import calculate_scaling_factors
from watertap.property_models import seawater_prop_pack
from watertap.examples.flowsheets.full_treatment_train.model_components import seawater_salt_prop_pack, seawater_ion_prop_pack
from watertap.examples.flowsheets.full_treatment_train.model_components.eNRTL import entrl_config_FpcTP
from watertap.examples.flowsheets.full_treatment_train.util import solve_with_user_scaling


def build_prop(m, base='TDS'):
    """
    Builds a property package for the specified base. Includes default scaling.
    Bases include: 'TDS', 'ion', 'salt'.
    """
    if base == 'TDS':
        m.fs.prop_TDS = seawater_prop_pack.SeawaterParameterBlock()

        m.fs.prop_TDS.set_default_scaling('flow_mass_phase_comp', 1, index=('Liq', 'H2O'))
        m.fs.prop_TDS.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'TDS'))

    elif base == 'ion':
        m.fs.prop_ion = seawater_ion_prop_pack.PropParameterBlock()

        m.fs.prop_ion.set_default_scaling('flow_mass_phase_comp', 1, index=('Liq', 'H2O'))
        m.fs.prop_ion.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'Na'))
        m.fs.prop_ion.set_default_scaling('flow_mass_phase_comp', 1e4, index=('Liq', 'Ca'))
        m.fs.prop_ion.set_default_scaling('flow_mass_phase_comp', 1e3, index=('Liq', 'Mg'))
        m.fs.prop_ion.set_default_scaling('flow_mass_phase_comp', 1e3, index=('Liq', 'SO4'))
        m.fs.prop_ion.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'Cl'))

    elif base == 'salt':
        m.fs.prop_salt = seawater_salt_prop_pack.PropParameterBlock()

        m.fs.prop_salt.set_default_scaling('flow_mass_phase_comp', 1, index=('Liq', 'H2O'))
        m.fs.prop_salt.set_default_scaling('flow_mass_phase_comp', 1e2, index=('Liq', 'NaCl'))
        m.fs.prop_salt.set_default_scaling('flow_mass_phase_comp', 1e3, index=('Liq', 'CaSO4'))
        m.fs.prop_salt.set_default_scaling('flow_mass_phase_comp', 1e3, index=('Liq', 'MgSO4'))
        m.fs.prop_salt.set_default_scaling('flow_mass_phase_comp', 1e3, index=('Liq', 'MgCl2'))

    elif base == 'eNRTL':
        m.fs.prop_eNRTL = GenericParameterBlock(default=entrl_config_FpcTP.configuration)

        # default scaling in config file

    else:
        raise ValueError('Unexpected property base {base} provided to build_prop'
                         ''.format(base=base))


def get_prop(m, base='TDS'):
    if base == 'TDS':
        prop = m.fs.prop_TDS
    elif base == 'ion':
        prop = m.fs.prop_ion
    elif base == 'salt':
        prop = m.fs.prop_salt
    elif base == 'eNRTL':
        prop = m.fs.prop_eNRTL
    else:
        raise ValueError('Unexpected property base {base} for get_prop'
                         ''.format(base=base))
    return prop



def specify_feed(sb, base='TDS'):
    """
    Fixes the state variables on the stateblock to the base seawater composition for
    the specified base. Bases include: 'TDS', 'ion', 'salt'.
    """
    sb.pressure.fix(101325)
    sb.temperature.fix(298.15)

    feed_flow_mass = 1
    if base == 'TDS':
        feed_mass_frac_TDS = 0.035
        sb.flow_mass_phase_comp['Liq', 'TDS'].fix(feed_flow_mass * feed_mass_frac_TDS)
        sb.flow_mass_phase_comp['Liq', 'H2O'].fix(feed_flow_mass * (1 - feed_mass_frac_TDS))

    elif base == 'ion':
        feed_mass_frac = {'Na': 11122e-6,
                          'Ca': 382e-6,
                          'Mg': 1394e-6,
                          'SO4': 2136e-6,
                          'Cl': 20316.88e-6}
        sb.flow_mass_phase_comp['Liq', 'H2O'].fix(
            feed_flow_mass * (1 - sum(x for x in feed_mass_frac.values())))
        for j in feed_mass_frac:
            sb.flow_mass_phase_comp['Liq', j].fix(feed_flow_mass * feed_mass_frac[j])

    elif base == 'salt':
        feed_mass_frac = {'NaCl': 2.827e-2,
                          'CaSO4': 1.298e-3,
                          'MgSO4': 1.529e-3,
                          'MgCl2': 4.251e-3,
                          'H2O': 0.9647}
        for s in feed_mass_frac:
            sb.flow_mass_phase_comp['Liq', s].fix(feed_flow_mass * feed_mass_frac[s])

    else:
        raise ValueError('Unexpected property base {base} provided to specify_feed'
                         ''.format(base=base))


def solve_specify_feed(base):
    # build state block
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    build_prop(m, base=base)
    if base == 'TDS':
        m.fs.stream = m.fs.prop_TDS.build_state_block([0], default={})
    elif base == 'ion':
        m.fs.stream = m.fs.prop_ion.build_state_block([0], default={})
    elif base == 'salt':
        m.fs.stream = m.fs.prop_salt.build_state_block([0], default={})
    specify_feed(m.fs.stream[0], base=base)

    m.fs.stream[0].mass_frac_phase_comp  # touch a variable to have a model with at least one constraint

    # scale
    calculate_scaling_factors(m.fs)
    # solve
    solve_with_user_scaling(m)
    # display
    m.fs.stream.display()

    return m


if __name__ == "__main__":
    solve_specify_feed('TDS')
    solve_specify_feed('ion')
    solve_specify_feed('salt')
